# Licensed under a 3-clause BSD style license - see LICENSE.rst

import copy
from functools import reduce
import operator
import os
from warnings import warn

import astropy.units as u
import astropy.io.fits as fits
import numpy as np
import numpy.ma as ma
import synphot as syn

from . import config as cf
from .constants import PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS
from .custom_exceptions import EngineOutputError, EngineInputError
from . import debug_utils
from .detector import H2RG, H4RG, SiAs, CCD, H1R, MAMA, XDL
from .instrument import check_warnings
from .io_utils import read_json
from .hst import HSTInstrument, COS
from .pandeia_warnings import report_warning_messages as warning_messages
from .projection import ProjectionFactory
from .utils import recursive_subclasses

default_refdata_directory = cf.default_refdata_directory

def ReportFactory(calc_input, signal_list, noise_list, saturation_list, extracted, warnings, sub_str=''):


    """
    Function to take strategy and build/return a configured instance of the appropriate report. For ease of
    identifying and making targeted changes to a single strategy when necessary, we define subclasses for
    every strategy.

    Parameters
    ----------
    calc_input: dict
        The input calculation dictionary
    signal_list: list
        List of all the signal objects (one per dither)
    noise_list: list
        List of all the noise objects (one per dither)
    saturation_list: list
        List of all the saturation objects (one per dither)
    extracted: dict
        Dictionary of all the extracted observation products
    warnings: dict
        Dictionary of warning messages produced by the code
    sub_str: string
        String for tagging warnings from different subreports

    Returns
    -------
    cls: Report class object
        Properly configured Report class for the given strategy
    """

    method = calc_input['strategy']['method']

    # we need to run the API checks here after merging in the report-specific defaults
    #if webapp:
    #    self._api_checks(all_config)

    types = recursive_subclasses(Report)
    methods = [t.__name__.lower().replace('report', '') for t in types]
    type_map = dict(list(zip(methods, types)))

    if method not in methods:
        msg = "Unsupported or not yet implemented strategy: %s" % method
        raise EngineInputError(value=msg)
    else:
        cls = type_map[method](calc_input, signal_list, noise_list, saturation_list, extracted, warnings, sub_str)
        return cls


class Report:

    """
    This class contains functionality for generating reports and data products from the results
    of a Pandeia calculation.

    Parameters
    ----------
    calc_input: dict
        The engine API format dict used to configure the calculation
    signal_list : list of etc3D.DetectorSignal instances
        The calculated signal on the detector
    noise_list : list of etc3D.DetectorNoise instances
        The calculated noise on the detector
    extracted : dict
        The extracted data product generated by the Strategy
    warnings: dict
        Dict of warnings produced and collected through the course of the calculation
    sub_str: str
        String for tagging warnings from different subreports
    """

    def __init__(self, calc_input, signal_list, noise_list, saturation_list, extracted_list, warnings, sub_str):
        # For single pointing calculations, the signal and noise lists have one element.
        # However, handling calculations that involve multiple pointings is more tricky.
        # We want to grab one of the on_target calculations to get the relevant target
        # information for the report.
        if not isinstance(signal_list, list):
            raise EngineOutputError(value="Must provide Report signals and noises as lists.")

        self.r = dict()

        self.warnings = warnings
        # We use the calculation inputs for a few things.
        self.input = calc_input

        self.r['sub_reports'] = []

        if len(signal_list) == 1:
            self.signal = signal_list[0]
            self.noise = noise_list[0]
            self.saturation = saturation_list[0]
            self.extracted = extracted_list[0]
        elif len(signal_list) > 1:
            # Fish out the list of which ones are targets.
            """
            on_target = calc_input['strategy']['on_target']  # True if target, False otherwise
            target = [i for i, v in enumerate(on_target) if v][0]  # get first target
            """
            # The above method does not work in general. Assume the first plane is a target and
            # refactor when implementing dithers in general.
            target = 0

            self.signal = signal_list[target]
            self.noise = noise_list[target]
            self.saturation = saturation_list[target]
            self.extracted = extracted_list[target]
            # the zeroth element in the extracted_list is actually the combined results (where they exist) followed by
            # the same planes that are in the signal, noise, and saturation lists. We need to remove it so they line up.
            # if the extracted results are not from a multi-pointing mode, they are identical to the zeroth elements of
            # the signal, noise, and saturation arrays.
            extracted_list.pop(0)  # we want the subreports to have the individual subreport extractions
            subreports = range(len(signal_list))
            for signal, noise, saturation, extracted, subreport in zip(signal_list, noise_list, saturation_list, extracted_list, subreports):
                s = ReportFactory(calc_input, [signal], [noise], [saturation], [extracted], warnings, '{0:d}'.format(subreport+1))
                self.r['sub_reports'].append(s.as_dict())

        self.sub_str = sub_str

        # Get the full Signal and Noise objects for inspection.
        self.signals = signal_list
        self.noises = noise_list
        self.saturations = saturation_list
        self.extracts = extracted_list

        self.tel = self.signal.current_instrument.telescope.tel_name
        self.inst = self.signal.current_instrument.inst_name
        self.mode = self.signal.current_instrument.mode
        self.strategy = self.input['strategy']['method']

        self.wave_index = 0
        self.projection_type = self.signal.projection_type
        self.projection = ProjectionFactory(self.projection_type)
        self._projection()

        # make a deep copy of the input so that nothing else can modify it later on
        self.r['input'] = copy.deepcopy(self.input)
        # initializing the rest of these as empty dictionaries to be filled
        self.r['1d'] = {}
        self.r['2d'] = {}
        self.r['3d'] = {}
        self.r['scalar'] = {}
        self.r['information'] = {}
        self.r['transform'] = {}

        # now fill the above dictionary sections with data
        self._information()
        self._transform()
        self._scalar()
        self._1d()
        self._debugarrays()
        self._2d()
        self._3d()

        if isinstance(self.signal.current_instrument, HSTInstrument):
            self.r['bop'] = {}
            self._bop()

            # Some HST instruments report in counts/s as opposed to electrons/s. Check the values accordingly.
            self.r["scalar"]["is_counts"] = False           
            if hasattr(self.signal.current_instrument.the_detector, "report_counts"):
                if self.signal.current_instrument.the_detector.report_counts:
                    self.r['scalar']['is_counts'] = True
                    self.convert_electrons_to_counts(calc_input)
            self.r['scalar']['obsmode'] = self.signal.current_instrument.obsmode

            if self.signal.current_instrument.instrument["mode"] == "ccd_imagingacq":
                self.r['scalar']['extracted_total_flux'] = self.r['scalar']['extracted_flux'] * self.signal.the_detector.exposure_spec.exposure_time

            # specially needed for WFC3 and STIS
            # the warning test that needs this checks per-independent-exposure.
            area = self.r['scalar']['background_area']
            if not isinstance(area, (float, int)): # Will be None if there's no background subtraction region
                area = self.r['scalar']['extraction_area']
            self.r['scalar']['background_sky_count_pixel'] = self.r['scalar']['background_sky'] * self.signal.the_detector.exposure_spec.exposure_time / area
            # ACS Ramp needs this
            pixscl_y, pixscl_x = self.signal.current_instrument.get_pix_size()
            self.r['scalar']['background_total_area'] = self.r['scalar']['background_total'] / area / (pixscl_y * pixscl_x)

            if 'brightest_pixel_rate' in self.r['bop']:
                bright_detector = max(zip(self.r['bop']['brightest_pixel_rate'].values(), self.r['bop']['brightest_pixel_rate'].keys()))[1]
                # get the brightest pixel per read in to the main report
                #self.r['scalar']['brightest_pixel'] = self.r['bop']['brightest_pixel_rate'][bright_detector][0]
                # get the brightest pixel count in to the main report
                self.r['scalar']['brightest_pixel_count'] = self.r['bop']['brightest_pixel_rate'][bright_detector][0] * self.signal.the_detector.exposure_spec.exposure_time

                # Copy brightest pixel wavelength, brightest pixel rate and detector total rate into the main report as well.
                self.r['scalar']['brightest_pixel_wavelength'] = self.r['bop']['brightest_pixel_rate'][bright_detector][1]
                self.r['scalar']['brightest_pixel_rate'] = self.r['bop']['brightest_pixel_rate'][bright_detector][0]
                self.r['scalar']['detector_total_rate'] = self.r['bop']['max_total_detector_rate']


            if isinstance(self.signal.current_instrument, COS):
                if not self.signal.single_point_source:
                    key="cos_extended_multiple"
                    self.warnings[key] = warning_messages[key]

            self._warnings()

            self._check_output_warnings(signal_list[0].current_instrument.warning_list_output)
            self.r["bop"]["warnings"] = {}

            if 'imag' in self.strategy:
                if len(self.input['scene']) > 1 or \
                   len(self.input['scene']) == 1 and (self.input['scene'][0]['position']['x_offset'] != 0.0 or self.input['scene'][0]['position']['y_offset'] != 0.0):
                    # BOP calculations are only accurate for scenes with a single, centered source.
                    # Otherwise, display a BOP warning.
                    key = 'bop_single_centered_source'
                    self.r['bop']['warnings'][key] = warning_messages[key]

            self._check_output_warnings_bop(self.signal.current_instrument.warning_list_bop)

            if len(self.r["bop"]["warnings"]) > 0:
                msg = "\033[1;31mWARNING: \033[1;33mThis calculation has critical Brightness Limit warnings. Check ['bop']['warnings'] dict in the result.\033[0m"
                warn(msg, FutureWarning)
        else:
            self._warnings()
            # And now, things that should NOT be done for HST
            # We do not want to display filter leak warnings for anyone but HST.
            warnings = self.r["warnings"]
            # we can't remove from a dict while we loop over it, though; so let's just rebuild it.
            self.r["warnings"] = {}
            for warning in warnings:
                if "filter_leak" not in warning:
                    self.r["warnings"][warning] = warnings[warning]

        # this must be run last, after the rest of the report has been filled.
        self._web_report()

    def _check_output_warnings(self, warning_list_output, sub_word=None, sub_str=''):

        # merge together input and results dictionaries so the check_warnings
        # function can find everything it needs.
        merged_dict = copy.deepcopy(self.input)
        merged_dict.update(self.r)

        # Each thing we want to be able to get the value of in the merged dictionary
        # should be listed here. The "value" part is a list of nested keys to get the actual value.
        #
        # Note that we can get in trouble when a parameter has a name that is part of a message
        # string substitution token. For example: if we name our variable 'detector_total_rate'
        # and add a reference to it to a message string, as $detector_total_rate, the string
        # substitution code will replace it with the detector name (if the detector name is one of
        # the selection parameters), resulting in the string "FUVMAMA_total_rate" showing up in
        # the output (example taken from STIS).
        value_lookup = {
            'instrument': ['configuration', 'instrument', 'instrument'],
            'mode': ['configuration', 'instrument', 'mode'],
            'detector': ['configuration', 'instrument', 'detector'],
            'disperser': ['configuration', 'instrument', 'disperser'],
            'gain': ['configuration', 'detector', 'gain'],
            'bin_spatial': ['configuration', 'detector', 'bin_spatial'],
            'bin_dispersion': ['configuration', 'detector', 'bin_dispersion'],
            'reference_wavelength': ['strategy', 'reference_wavelength'],
            'target_total_rate': ['scalar', 'extracted_flux'],
            'time': ['scalar', 'saturation_time'],
            'scan_rate': ['scalar', 'scan_rate'],
            'background_sky_count_pixel': ['scalar', 'background_sky_count_pixel'],
            'brightest_pixel_count': ['bop', 'max_bright_pixel_count'],
            'brightest_pixel_rate': ['bop', 'max_bright_pixel_rate'],
            #
            #TODO placeholder for after JETC-2421 gets merged
            'brightest_segment_rate': ['bop', 'max_bright_pixel_rate'],
            #
            'det_total_rate': ['bop', 'max_total_detector_rate'],
            'total_leak': ['scalar', 'total_leak'],
            'buffer_time': ['scalar', 'buffer_time'],
            'sn': ['scalar', 'sn'],
            'n_partial_saturated': ['scalar', 'n_partial_saturated'],
            'n_full_saturated': ['scalar', 'n_full_saturated'],
            'n_cube_partial_saturated': ['scalar', 'n_cube_partial_saturated'],
            'n_cube_full_saturated': ['scalar', 'n_cube_full_saturated'],
            'n_total_saturated': ["scalar", "n_total_saturated"],
            'max_saturated_pixels': ["scalar", "max_saturated_pixels"],
            'min_snr_threshold': ["scalar", "min_snr_threshold"]
        }

        check_warnings(merged_dict, self.warnings, warning_list_output, value_lookup, sub_word=sub_word, sub_str=sub_str)

    def _check_output_warnings_bop(self, warning_list_output_bop, sub_word=None, sub_str=''):

        # merge together input and results dictionaries so the check_warnings
        # function can find everything it needs.
        merged_dict = copy.deepcopy(self.input)
        merged_dict.update(self.r)

        # Each thing we want to be able to get the value of in the merged dictionary
        # should be listed here. The "value" part is a list of nested keys to get the actual value.
        #
        # Note that we can get in trouble when a parameter has a name that is part of a message
        # string substitution token. For example: if we name our variable 'detector_total_rate'
        # and add a reference to it to a message string, as $detector_total_rate, the string
        # substitution code will replace it with the detector name (if the detector name is one of
        # the selection parameters), resulting in tye string "FUVMAMA_total_rate" showing up in
        # the output (example taken from STIS).
        value_lookup = {
            'instrument': ['configuration', 'instrument', 'instrument'],
            'mode': ['configuration', 'instrument', 'mode'],
            'detector': ['configuration', 'instrument', 'detector'],
            'disperser': ['configuration', 'instrument', 'disperser'],
            'gain': ['configuration', 'detector', 'gain'],
            'bin_spatial': ['configuration', 'detector', 'bin_spatial'],
            'bin_dispersion': ['configuration', 'detector', 'bin_dispersion'],
            'reference_wavelength': ['strategy', 'reference_wavelength'],
            'target_total_rate': ['scalar', 'extracted_flux'],
            'time': ['scalar', 'saturation_time'],
            'scan_rate': ['scalar', 'scan_rate'],
            'background_sky_count_pixel': ['scalar', 'background_sky_count_pixel'],
            'brightest_pixel_count': ['bop', 'max_bright_pixel_count'],
            'brightest_pixel_rate': ['bop', 'max_bright_pixel_rate'],
            #
            #TODO placeholder for after JETC-2421 gets merged
            'brightest_segment_rate': ['bop', 'max_bright_pixel_rate'],
            #
            'det_total_rate': ['bop', 'max_total_detector_rate'],
            'total_leak': ['scalar', 'total_leak'],
            'buffer_time': ['scalar', 'buffer_time'],
        }

        check_warnings(merged_dict, self.r["bop"]["warnings"], warning_list_output_bop, value_lookup, sub_word=sub_word, sub_str=sub_str)

    def as_dict(self):
        """
        Functions to return the internally-created dictionary.
        It is possible to access the report dictionary attribute directly, but for backwards compatibility we've
        kept this function.

        Returns
        -------
        self.r : dict
            output dictionary of results
        """
        return self.r

    def as_fits(self):
        """
        Returns a dictionary of fits headers.
        This cannot be a separate function because it requires several attributes of the class.

        Returns
        -------
        self.f : dict
            output dictionary of fits tables
        """

        sub_reports = []
        if len(self.r['sub_reports']) > 1:
            for self.report in range(len(self.r['sub_reports'])):
                self.s = self.r['sub_reports'][self.report]
                self.f = dict()
                self.f['3d'] = {}
                self.f['2d'] = {}
                self.f['1d'] = {}
                self._3dfits()
                self._2dfits()
                self._1dfits()
                sub_reports.append(self.f)

        self.report = 0
        self.s = self.r
        self.f = dict()
        self.f['3d'] = {}
        self.f['2d'] = {}
        self.f['1d'] = {}
        self._3dfits()
        self._2dfits()
        self._1dfits()
        self.f['sub_reports'] = sub_reports

        return self.f

    def _set_wave(self):
        """
        Find and set the wavelength to extract scalar values

        Parameters
        ----------
        wref: float
            Wavelength desired for the slice (in microns)
        """
        if len(self.wave_pix) == 1:
            self.rw = self.wave_pix[0]
        else:
            self.rw = (np.min(self.wave_pix) + np.max(self.wave_pix))/2.0
            if "reference_wavelength" in self.input["strategy"]:
                wref = self.input["strategy"]["reference_wavelength"]
                if wref is not None:
                    if wref >= self.wave_pix.min() and wref <= self.wave_pix.max():
                        self.rw = wref
                    else:
                        self.warnings["bad_waveref"] = warning_messages["bad_waveref"].format(wref,self.wave_pix.min(),
                                                                                        self.wave_pix.max(),
                                                                                        self.rw)
                # Not setting a reference wavelength is legitimate
            else:
                self.warnings["no_waveref"] = warning_messages["no_waveref"].format(self.rw)
        if not hasattr(self.signal.the_detector, "pixels_per_resel"):
            wave_index = (np.abs(self.wave_pix - self.rw)).argmin()
            self.rw = self.wave_pix[wave_index]

    def _3dfits(self):
        """
        Format 3D outputs into fits headers.
        """
        for k in ['flux', 'flux_plus_background']:
            self.f['3d'][k] = []
            for i in range(len(self.s['3d'][k])):
                # the cube data coming out of pandeia.engine is ordered wavelength,x,y which is
                # backwards from the numpy convention of ordering axes from slowest to fastest.
                # it is set up this way because we need to broadcast 1D vectors (e.g. throughput
                # vs. wavelength) onto the 3D cubes. for this to work, wavelength must be the 3rd
                # axis. so flip the dimensions of the cube and then flip the Y axis to get the expected
                # orientation in the FITS data as viewed in DS9 or the like.
                o = fits.PrimaryHDU(np.moveaxis(self.s['3d'][k][i],2,0)[:, ::-1, :])
                # it should not be different, but pull the wcs info from the correct signal
                tbhdu, header = self.signals[self.report].cube_wcs_info()
                o.header.update(header)
                self.f['3d'][k].append(fits.HDUList([o, tbhdu]))

    def _2dfits(self):
        """
        Format 2D outputs into fits headers.
        """
        # this is set by _projection, and should be the same for all subreports of a single calculation
        detector_header = self.pix_grid.wcs_info()
        for k in self.s['2d']:
            # flip the Y axis to get the FITS data to look the
            # same in DS9 as in matplotlib.
            o = fits.PrimaryHDU(self.s['2d'][k][::-1, :])
            o.header.update(detector_header)
            self.f['2d'][k] = o

    def _1dfits(self):
        """
        Format 1D outputs into fits headers.
        """

        # These two are special one-array lists
        for k in ['wave_calc', 'wave_pix']:
            tbhdu = fits.BinTableHDU.from_columns([
                fits.Column(name='WAVELENGTH',
                            unit='um',
                            format="1D",
                            array=self.s['1d'][k])
            ])
            tbhdu.name = k
            self.f['1d'][k] = fits.HDUList([tbhdu])

        # now go through the rest
        # note that we're removing them from a list of dictionary keys, not removing them from the actual dictionary.
        results1d = list(self.s['1d'].keys())
        results1d.remove('wave_calc')
        results1d.remove('wave_pix')
        for k in results1d:
            tbhdu = fits.BinTableHDU.from_columns([
                fits.Column(name='WAVELENGTH',
                            unit='um',
                            format="1D",
                            array=self.s['1d'][k][0]),
                fits.Column(name=k,
                            format="1D",
                            array=self.s['1d'][k][1])
            ])
            tbhdu.name = k
            self.f['1d'][k] = fits.HDUList([tbhdu])

    def _projection(self):
        """
        Run the appropriate projection class. This is often overridden by strategies to include additional code.
        """
        self.grid, self.wave_pix, self.pix_grid = self.projection._project(self.signal, self.extracted)

    def _transform(self):
        """
        Spatial and wavelength information taken from the coordinate transform products
        """

        # set up the coordinate transform information that describes the image axes
        # for the 2D and 3D data.  first, get the spatial information from self.Grid

        self.r['transform'] = self.pix_grid.as_dict()

        # the wavelength sampling of the input spectrum is different
        # then the sampling by the detector. the model cubes that are
        # used to make the calculation use the input sampling. the detector
        # plane outputs for spectroscopy and IFU use the detector sampling

        # ## input model wavelength sampling ## #
        self.r['transform'].update(self.signal.spectral_model_transform())

        # ## detector plane wavelength sampling ## #
        self.r['transform'].update(self.signal.spectral_detector_transform())

        # the spectral mapping can be shifted by strategies (e.g. slitless). this is
        # reflected in wave_pix so update from there.
        self.r['transform']['wave_det_min'] = self.wave_pix.min()
        self.r['transform']['wave_det_max'] = self.wave_pix.max()
        self.r['transform']['wave_det_refval'] = self.wave_pix[0]

    def _scalar(self):
        """
        Scalar data products

        Note that this method allows for certain variables to be unspecified in the inputs, on a case by case basis.
        This permissiveness is an intentional engine feature - parameters that are not given in the inputs are to be
        filled in by (sensible, astronomer-verified) defaults, and should not cause crashes. #3737
        """

        # Get the scalar wavelength to extract
        self._set_wave()

        # get areas in pixels of source and background regions
        if isinstance(self.extracted["extraction_area"], (list, np.ndarray)):
            self.r['scalar']['extraction_area'] = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracted["extraction_area"],  self.rw)
            self.extraction_area = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracted["extraction_area"],  self.rw)
        else:
            self.r['scalar']['extraction_area'] = self.extracted['extraction_area']
            self.extraction_area = self.extracted['extraction_area']

        self.r['scalar']['background_area'] = self.extracted['background_area']
        self.background_area = self.extracted['background_area']

        self.r['scalar']['fraction_saturation'] = self.extracted['saturation_products']['fraction_saturation']
        self.r['scalar']['sat_ngroups'] = np.min(self.extracted['detector_ngroups'])
        self.r['scalar']['brightest_pixel'] = self.signal.brightest_pixel

        # buffer time - must handle segmented and non-segmented detectors.
        if hasattr(self.signal.the_detector, "half_buffer"):
            if isinstance(self.signal.detector_total_rate, dict) and len(self.signal.detector_total_rate) > 1:
                result = {}
                detector_total_rate = 0

                for segment in self.signal.detector_total_rate:
                    result[segment] = self.signal.the_detector.half_buffer / self.signal.detector_total_rate[segment]
                    detector_total_rate += self.signal.detector_total_rate[segment]

                self.r['scalar']['segmented_buffer_time'] = result
                self.r['scalar']['buffer_time'] = self.signal.the_detector.half_buffer / detector_total_rate
            else:
                key = list(self.signal.detector_total_rate.keys())[0]
                self.r['scalar']['buffer_time'] = self.signal.the_detector.half_buffer / \
                                                  self.signal.detector_total_rate[key]

        # segmented detectors report brightest pixel per segment
        if hasattr(self.signal, 'brightest_pixel_rate'):
            if isinstance(self.signal.brightest_pixel_rate, dict)  and len(self.signal.brightest_pixel_rate) > 1:
                self.r['scalar']['segmented_brightest_pixel_rate'] = self.signal.brightest_pixel_rate

        # segmented detectors report total rate and buffer time per segment
        if hasattr(self.signal, 'detector_total_rate'):
            if isinstance(self.signal.detector_total_rate, dict) and len(self.signal.detector_total_rate) > 1:
                self.r['scalar']['brightest_segment_rate'] = max(self.signal.detector_total_rate.values())
                for key in self.signal.detector_total_rate:
                    self.r['scalar']['segment_' + key + '_total_rate'] = self.signal.detector_total_rate[key]

        # put some instrument and strategy configuration info into the scalar report
        self.r['scalar']['filter'] = self.input['configuration']['instrument']['filter']
        self.r['scalar']['disperser'] = self.input['configuration']['instrument']['disperser']
        if "cenwave" in self.input['configuration']['instrument']:
            self.r['scalar']['cenwave'] = self.input['configuration']['instrument']['cenwave']
        if "slit" in self.input['configuration']['instrument']:
            self.r['scalar']['slit'] = self.input['configuration']['instrument']['slit']
        if "postflash" in self.input['configuration']['detector']:
            self.r['scalar']['postflash'] = self.input['configuration']['detector']['postflash']

        # target_xy is not found in SOSS or MSAFullApPhot, and may not be found in future strategies as well.
        # If it's not there, assume the target position is the center of the FOV, i.e. (0.0, 0.0).
        if 'target_xy' in self.input['strategy']:
            self.r['scalar']['x_offset'] = self.input['strategy']['target_xy'][0]
            self.r['scalar']['y_offset'] = self.input['strategy']['target_xy'][1]
        else:
            self.r['scalar']['x_offset'] = 0.0
            self.r['scalar']['y_offset'] = 0.0

        # Not all strategies define an aperture size either. SOSS doesn't because the size is defined by an
        # the grism. TAPhot and TACentroid do not require it to be specified, as it is fixed and cannot be changed
        # by users. Optimal extraction strategies likely will not use this, either, once implemented.
        if 'aperture_size' in self.input['strategy']:
            self.r['scalar']['aperture_size'] = self.input['strategy']['aperture_size']

            if "aperture_sizes_arcsec" in self.extracted:
                # If the aperture is defined by the enclosed energy percentage, we need to get the aperture_size
                #  in arcseconds from the extracted quantities.
                self.r['scalar']['aperture_size_percent'] = self.input['strategy']['aperture_size']
                self.r['scalar']['aperture_size'] = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracted["aperture_sizes_arcsec"],  self.rw)
        else:
            self.r['scalar']['aperture_size'] = "n/a"

    def _bop(self):
        """
        Compute HST's Bright Object Protection (Health and Safety) quantities
        """

        # TODO  -  JETC-2484
        # This is a temporary bail-out code that prevents bop quantities to be calculated
        # when the flux cube contains nans. Otherwise, the code downstream that expects
        # bop values, crashes. To bem removed once the ticket above gets implemented.
        if len(self.signal.brightest_pixel_rate.values()) == 0:
            return

        self.r['bop']['brightest_pixel_rate'] = self.signal.brightest_pixel_rate
        self.r['bop']['detector_total_rate'] = self.signal.detector_total_rate

        # BOP brightest pixel rate and detector total rate are provided for more than one segments
        # find the highest per-segment value and add BOP warning based on it

        detector = self.input['configuration']['instrument']['detector']
        # max brightest pixel rate from the dict with tuples for rate and wavelength at which it occurs for every chip
        time = self.signal.the_detector.exposure_spec.exposure_time # for a single exposure, taking nsplit into account
        max_bright_pixel_rate = max(self.signal.brightest_pixel_rate.values(), key = operator.itemgetter(0))[0]
        self.r['bop']['max_bright_pixel_rate'] = max_bright_pixel_rate
        self.r['bop']['max_bright_pixel_count'] = max_bright_pixel_rate * time
        self.r['bop']['max_total_detector_rate'] = max(self.signal.detector_total_rate.values())

    def _3d(self):
        """
        Compute 3D data products
        """

        # this is the data cube of the input signal (not converted to photons, with no telescope throughputs)
        self.flux = self.signal.flux_cube_list
        self.flux_plus_bg = self.signal.flux_plus_bg_list
        self.r['3d']['flux'] = self.flux

        # this is the data cube of the input signal plus background
        self.r['3d']['flux_plus_background'] = self.flux_plus_bg

    def _2d(self):
        """
        Compute 2D data products. Though some strategies override this, these are mostly projection-dependent
        """
        resultdict,self.bg_pix = self.projection._2d(self.extracted, self.saturation, self.signal)
        self.r['2d'].update(resultdict)

    def _1d(self):
        """
        Compute 1D data products, and the scalar values that are extracted from the 1D products.

        NOTE: We do not do any NaN checking here.  This may cause certain values like SNR
        to be reported as NaN in 1d, but as 0 in scalars.
        """

        # this is the wavelength sampling of the calculation
        self.wave = self.signal.wave

        self.r['1d']['wave_pix'] = self.wave_pix
        self.r['1d']['wave_calc'] = self.wave

        # Target spectrum
        self.r['1d']['target'] = [self.wave, self.signal.total_flux]

        # Transmission/focal plane rate
        self.r['1d']['fp'] = [self.wave, self.signal.total_rate[0]]

        # input background
        self.r['1d']['bg'] = [self.wave, self.signal.background.MJy_sr]

        # Background rate
        self.r['1d']['bg_rate'] = [self.wave, self.signal.bg_fp_rate]

        # Values calculated by the strategy
        sn = self.extracted['extracted_flux'] / self.extracted['extracted_noise']

        self.r['1d']['sn'] = [self.wave_pix, sn]
        self.r['1d']['extracted_noise'] = [self.wave_pix, self.extracted['extracted_noise']]
        self.r['1d']['extracted_flux'] = [self.wave_pix, self.extracted['extracted_flux']]
        self.r['1d']['extracted_flux_plus_bg'] = [self.wave_pix, self.extracted['extracted_flux_plus_bg']]

        self.r['1d']['total_flux'] = [self.wave_pix, self.extracted['source_flux_in_fov']]

        self.r['1d']['extracted_bg_total'] = [self.wave_pix, self.extracted['extracted_bg_total']]
        self.r['1d']['extracted_bg_only'] = [self.wave_pix, self.extracted['extracted_bg_only']]
        self.r['1d']['extracted_contamination'] = [self.wave_pix, self.extracted['extracted_bg_total'] - self.extracted['extracted_bg_only']]
        self.r['1d']['n_partial_saturated'] = [self.wave_pix, self.extracted['saturation_products']['partial']]
        self.r['1d']['n_full_saturated'] = [self.wave_pix, self.extracted['saturation_products']['full']]

        self.curves = self.r['1d']

        # Values calculated by the strategy
        # CHECK: if this used dispersion binning, set resels to the minimum of (1, Detector.pixels_per_resel/dispersion_binning)
        if hasattr(self.signal.the_detector, "pixels_per_resel"):
            if hasattr(self.signal.the_detector, "bin_dispersion") and self.signal.the_detector.bin_dispersion != 1:
                self.signal.the_detector.pixels_per_resel = np.min(1, self.signal.the_detector.pixels_per_resel/self.signal.the_detector.bin_dispersion)
            self.r['1d']['extracted_noise_resel'] = self.signal.the_detector.to_resels_array(self.wave_pix, self.extracted['extracted_noise'])
            self.r['1d']['extracted_flux_resel'] = self.signal.the_detector.to_resels_array(self.wave_pix, self.extracted['extracted_flux'])
            self.r['1d']['extracted_flux_plus_bg_resel'] = self.signal.the_detector.to_resels_array(self.wave_pix, self.extracted['extracted_flux_plus_bg'])

            self.r['1d']['extracted_bg_total_resel'] = self.signal.the_detector.to_resels_array(self.wave_pix, self.extracted['extracted_bg_total'])
            self.r['1d']['extracted_bg_only_resel'] = self.signal.the_detector.to_resels_array(self.wave_pix, self.extracted['extracted_bg_only'])
            self.r['1d']['extracted_contamination_resel'] = self.signal.the_detector.to_resels_array(self.wave_pix, self.extracted['extracted_bg_total'] - self.extracted['extracted_bg_only'])
            self.r['1d']['n_partial_saturated_resel'] = self.signal.the_detector.to_resels_array(self.wave_pix, self.extracted['saturation_products']['partial'])
            self.r['1d']['n_full_saturated_resel'] = self.signal.the_detector.to_resels_array(self.wave_pix, self.extracted['saturation_products']['full'])

            sn = self.r['1d']['extracted_flux_resel'][1] / self.r['1d']['extracted_noise_resel'][1]
            self.r['1d']['sn_resel'] = [self.r['1d']['extracted_flux_resel'][0], sn]

        # Get the scalar wavelength to extract
        self._set_wave()

        signal = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracted['extracted_flux'], self.rw)
        noise = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracted['extracted_noise'], self.rw,
                                                          squared=True)
        self.r['scalar']['sn'] = signal/noise
        if not np.isfinite(self.r['scalar']['sn']):
            self.r['scalar']['sn'] = 0.0
        self.r['scalar']['extracted_flux'] = signal

        # if the SN is 0.0 (e.g. due to heavy saturation), we do not know what the variance is and set it to np.nan
        if np.abs(self.r['scalar']['sn']) > 0.0:
            self.r['scalar']['extracted_noise'] = noise
        else:
            self.r['scalar']['extracted_noise'] = np.nan

        self.r['scalar']['background_total'] = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracted['extracted_bg_total'], self.rw)
        self.r['scalar']['background_sky'] = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracted['extracted_bg_only'], self.rw)

        # only check for contamination if background level is above some epsilon value. otherwise round-off
        # errors can cause unstable test results.
        if np.abs(self.r['scalar']['background_total']) > 1.0e-9:
            self.r['scalar']['contamination'] = (self.r['scalar']['background_total'] - self.r['scalar']['background_sky']) / \
                                                self.r['scalar']['background_total']
        else:
            self.r['scalar']['contamination'] = 0.0

        self.r['scalar']['reference_wavelength'] = self.rw
        self.r['scalar']['background'] = self.signal.background.bg_spec(self.rw * PANDEIA_WAVEUNITS, flux_unit=u.MJy).value


    def _information(self):
        """
        Collect some additional information that is not the product of a calculation.
        """

        # get the ExposureSpecification.
        self.exposure_specification = self.signal.the_detector.exposure_spec

        self.r['information']['calc_type'] = self.signal.projection_type

        self.r['information']['exposure_specification'] = self.exposure_specification.__dict__

        # Some strategies like IFU and coronagraphy add dithers. Some are on-source, others are off-source.
        # This ensures that all on-source dithers are included in the "total_exposure_time", whereas
        # off-source time is only reported in "all_dithers_time".
        total_exposure_time = self.r['information']['exposure_specification']['total_exposure_time']
        if 'n_on_source' in self.extracted:
            self.r['scalar']['total_exposure_time'] = total_exposure_time * self.extracted['n_on_source']
            self.r['scalar']['all_dithers_time'] = total_exposure_time * self.extracted['n_total']
        else:
            self.r['scalar']['total_exposure_time'] = total_exposure_time
            self.r['scalar']['all_dithers_time'] = total_exposure_time

        # Report on measurement, exposure time (time for a single exposure) and saturation time.
        # The r['scalar']['exposure_time'] value reported back to the client is not used at this point.
        self.r['scalar']['exposure_time'] = self.r['information']['exposure_specification']['exposure_time']
        self.r['scalar']['measurement_time'] = self.r['information']['exposure_specification']['measurement_time']
        self.r['scalar']['saturation_time'] = self.r['information']['exposure_specification']['saturation_time']
        self.r['scalar']['total_integrations'] = self.r['information']['exposure_specification']['total_integrations']

        # Report the exposure duty cycle
        self.r['scalar']['duty_cycle'] = self.r['information']['exposure_specification']['duty_cycle']

        # get the pixel CR rate in events/s and convert to events/ramp
        self.r['scalar']['cr_ramp_rate'] = self.noise.pix_cr_rate * self.r['information']['exposure_specification'][
            'saturation_time']

    def _warnings(self, sub_words={'': None, '1':'Dither','2':'Dither', '3': 'Dither'}):
        """
        Collect warnings, and generate saturation warnings

        This module contains both the collected warnings from elsewhere in the program, and also the saturation
        warnings (which are, currently, the only warnings that may be different between subreports)

        Parameters
        -----------
        sub_words : dict
            dict of strings to use when differentiating warning subreports
        """

        # The only time the saturation warnings do NOT come from self.extracted['detector_saturation'] is with the
        # IFU modes, which override this method.
        sat_obj = self.extracted["detector_saturation"]
        self.r["warnings"] = self.warnings

        self.r["scalar"]["n_partial_saturated"] = (sat_obj == 1).sum()
        self.r["scalar"]["n_full_saturated"] = (sat_obj == 2).sum()

        if not isinstance(self.signal.the_detector, (MAMA, XDL)):
            self.r["scalar"]["fullwell"] = self.signal.the_detector.fullwell


        self._check_output_warnings(self.signal.current_instrument.warning_list_output, sub_word=sub_words[self.sub_str], sub_str=self.sub_str)

        # these properties were only needed for warning generation
        del self.r["scalar"]["n_partial_saturated"]
        del self.r["scalar"]["n_full_saturated"]

        if not isinstance(self.signal.the_detector, (MAMA, XDL)):
            del self.r["scalar"]["fullwell"]


    def _debugarrays(self):
        """
        Collect the debug information if it exists
        """
        if debug_utils.debugarrays and debug_utils.debugarrays.get_report():
            self.r['debugarrays'] = debug_utils.debugarrays.get_report()


    def convert_electrons_to_counts(self, calc_input):
        """
        This is an HST instrument and gain is present, so convert electrons to counts for the relevant keys.
        gain = electrons / counts --> counts = electrons / gain

        Outputs will be in counts/s
        """
        gain = float(calc_input['configuration']['detector'].get('gain',1.0))

        if 'background_total' in self.r['scalar']:
            self.r['scalar']['background_total'] = self.r['scalar']['background_total'] * gain
            self.r['scalar']['background_sky'] = self.r['scalar']['background_sky'] * gain
        if 'brightest_pixel' in self.r['scalar']:
            self.r['scalar']['brightest_pixel'] = self.r['scalar']['brightest_pixel'] * gain
        if 'extracted_flux' in self.r['scalar']:
            self.r['scalar']['extracted_flux'] = self.r['scalar']['extracted_flux'] * gain
        if 'extracted_noise' in self.r['scalar']:
            self.r['scalar']['extracted_noise'] = self.r['scalar']['extracted_noise'] * gain

        if 'brightest_pixel_rate' in self.r['bop']:
            for detector in self.r['bop']['brightest_pixel_rate']:
                self.r['bop']['brightest_pixel_rate'][detector][0] = self.r['bop']['brightest_pixel_rate'][detector][0] * gain

        if 'detector_total_rate' in self.r['bop']:
            for key, value in self.r['bop']['detector_total_rate'].items():
                self.r['bop']['detector_total_rate'][key] = self.r['bop']['detector_total_rate'][key] * gain

############ The pretty-printed web report #################

    def _web_report(self):
        """
        Create a pretty-formatted report for UI (or other) display

        This operates on a dictionary that defines what the output should look like, which is
        produced by _get_report_config.

        For the most part, the only part of the dictionary that needs to be changed between the engine definition
        and web UI output format is that the keys need to be replaced with their corresponding value.
        """
        report_config, inst_precision = self._get_report_config()
        # web_report is a list to preserve ordering
        self.r['web_report'] = []
        # loop through all the categories
        for category in report_config:
            # the items in each category are a list to preserve ordering
            category_dict = {"category": category, "items": []}
            # special keywords go here, and must be removed before we process rows
            if "color" in report_config[category]:
                category_dict["color"] = report_config[category]["color"]
                del report_config[category]["color"]

            # Now process the rows
            for item in range(len(report_config[category])):
                # start building the ending statement
                insertion = {}

                # 1. Get the name, which may be a dict whose choice depends on strategy
                if isinstance(report_config[category][item]["name"], dict):
                    if self.strategy in report_config[category][item]["name"]:
                        row_name = report_config[category][item]["name"][self.strategy]
                    else:
                        row_name = report_config[category][item]["name"]["default"]
                else:
                    row_name = report_config[category][item]["name"]
                insertion["name"] = row_name

                # 2. (optional) color formatting
                if "color" in report_config[category][item]:
                    insertion["color"] = report_config[category][item]["color"]

                # If there is a keylist we have a value(s) to extract
                if "key" in report_config[category][item]: # we will not always have keys for a value, so as to allow plain text to appear in the outputs.
                    precision = inst_precision
                    if "precision" in report_config[category][item]:
                        precision = report_config[category][item]["precision"]
                    # 3. Get the unit, which may be a dict whose choice depends on mode
                    if isinstance(report_config[category][item]["unit"], dict):
                        if self.mode in report_config[category][item]["unit"]:
                            row_unit = report_config[category][item]["unit"][self.mode]
                        else:
                            row_unit = report_config[category][item]["unit"]["default"]
                        # because we do occasionally reload the name from report_config object, we do need to change both of them
                    elif report_config[category][item]["unit"] == "set-by-software":
                        # Some items have units that depend on user inputs.  Set the units for those items here.
                        if report_config[category][item]["key"][0][1] == "aperture_size":
                            row_unit = self.input['strategy']['units']
                            if row_unit != 'arcsec':
                                row_unit += 's'
                    else:
                        row_unit = report_config[category][item]["unit"]
                    insertion["unit"] = row_unit

                    # 4. Now walk down the keylist to get the value
                    valuelist = []
                    # 4a. we will either have multiple keys per value (ex: target position) OR multiple values per key (ex: COS Spectroscopy)
                    if len(report_config[category][item]["key"]) > 1:
                        for keyset in report_config[category][item]["key"]:
                            val = self._key_walk(keyset, row_unit, precision)
                            valuelist.append(val.strip())
                        row_value = f"{', '.join(valuelist)}"
                    else:
                        row_value = self._key_walk(report_config[category][item]["key"][0], row_unit, precision)

                    # 4b. if the row_value is not a list, it's a single value to be added. 
                    if not isinstance(row_value,list):
                        insertion["value"] = row_value
                        insertion_list = [insertion]
                    else:
                        # If the row_value is a list, it will be a list of dicts with name, value pairs to differentiate detectors.
                        insertion_list = []
                        for single_value in row_value:
                            # Only the combined value will be formatted to allow us to insert the original name.
                            # This step is different because we have multiple outputs that each become their own line.

                            new_insertion = copy.deepcopy(insertion)
                            new_insertion["name"] = single_value["name"].format(row_name)
                            new_insertion["value"] = single_value["val"] 
                            new_insertion["indent"] = single_value["indent"]
                            insertion_list.append(new_insertion)
                else:
                    # No key, so just display the text.
                    insertion_list = [insertion]
                category_dict["items"].extend(insertion_list)
            self.r["web_report"].append(category_dict)


    def _get_report_config(self):
        """
        Set up formatting for this calculation
        """
        tel_config_path = os.path.join(default_refdata_directory, self.tel, "telescope", "report_config.json")
        inst_config_path = os.path.join(default_refdata_directory, self.tel, self.inst, "report_config.json")
        tel_config = read_json(tel_config_path, raise_except=True)
        inst_config = read_json(inst_config_path, raise_except=True)

        inst_precision = tel_config["precision"]
        if "precision" in inst_config:
            inst_precision = inst_config["precision"]
            del inst_config["precision"]
        report_config = {category: tel_config[category]['default'] for category in tel_config if category not in ["precision", "-delivered-for-version-"]}

        # Order of precedence: (on a section-by-section basis)
        # 1. Mode-specific section of the inst configuration
        # 2. Inst-specific config
        # 3. Telescope default
        #
        # Individual modes can remove entries with a special keyword
        for category in inst_config:
            if self.mode in inst_config[category]:
                report_config[category] = inst_config[category][self.mode]
                if "delete" in report_config[category]:
                    del report_config[category]
            elif "default" in inst_config[category]:
                report_config[category] = inst_config[category]["default"]
            # else use the telescope configuration

        return report_config, inst_precision

    def _key_walk(self, keys, unit, precision):
        """
        Actually walk through the report dictionary to find the value and format it.
        Some very basic processing can be done here.

        """
        # https://stackoverflow.com/questions/9320335/accessing-python-dict-with-multiple-key-lookup-string
        try:
            val = reduce(dict.get, keys, self.r)
        except TypeError: # Failures to find the key result in a TypeError, not a KeyError.
            val = None

        # Some unit conversions here
        if unit == "Angstroms" and val is not None:
            val *= 10000

        val = self._value_process(val, precision)
        return val

    def _value_process(self, val, precision, spacing=True):
        """
        Process a value into a string according to what it is
        """
        if isinstance(val, (float, np.floating)):
            if val != 0.0 and np.log(abs(val)) < -1*precision:
                # I put a format string in our format string so we can format our format
                val = format(val, f" .{precision}e")
            else:
                val = format(val, f" .{precision}f")
        elif isinstance(val, (int, np.signedinteger)):
            val = f"{val}{' '*(precision+1)}"
        elif isinstance(val, (list,tuple, np.ndarray)):
            val = f"({','.join([self._value_process(x, precision).strip() for x in val])})"
        elif isinstance(val, dict):
            valuelist = []
            outputlist = []
            for key in val:
                if isinstance(val[key], (list, tuple)):
                    # this is a bit of hard-coding, but currently the only time the dict contains a list, it's brightest_pixel_rate
                    outputlist.append({"name": f'{key}', "val": f"{self._value_process(val[key][0], precision)} ({self._value_process(val[key][1]*10000, precision).strip()} A)", "indent": True})
                    valuelist.append(val[key][0])
                else:
                    # name each individual output after the key (aka chip/stripe)
                    outputlist.append({"name": f'{key}', "val": self._value_process(val[key], precision), "indent": True})
                    valuelist.append(val[key])
            # if there's more than one value, find the maximum and add a name that will pick up the main name
            if len(valuelist) > 1:
                outputlist.insert(0, {"name": "{}", "val": self._value_process(max(valuelist), precision), "indent": False})
                val = outputlist
            # otherwise, process as if it's just one value
            else:
                val = self._value_process(max(valuelist), precision)
        elif val is None:
            val = "n/a"
        # otherwise, it's a string and we just return it
        return val

############ End pretty-printed web report #################

class CoronagraphyReport(Report):
    """
    This report class is specifically for Coronagraphy (and its derivatives)
    Parameters are read in through Report.__init__

    """

    def _1d(self):
        """
        Coronagraphy-specific contrast quantities
        """
        Report._1d(self)
        # Coronagraphy always has contrast curves
        self.r['1d']['contrast'] = self.extracted['contrast_curve']
        self.curves['contrast'] = self.extracted['contrast_curve']

    def _1dfits(self):
        Report._1dfits(self)
        # contrast is NOT like the other 1D products and must be handled separately (overwritten).
        tbhdu = fits.BinTableHDU.from_columns([
            fits.Column(name='SEPARATION',
                        unit='arcsec',
                        format="1D",
                        array=self.s['1d']['contrast'][0]),
            fits.Column(name="CONTRAST",
                        format="1D",
                        array=self.s['1d']['contrast'][1])
        ])
        tbhdu.name = 'contrast'
        self.f['1d']['contrast'] = fits.HDUList([tbhdu])

    def _2d(self):
        """
        Coronagraphy-specific 2D calculations.

        This does not call any of the usual _2d() infrastructure (in projection.py classes) because Coronagraphy
        2D planes are different from any other mode. The coronagraphy strategy produces (and uniquely, uses) a
        detector_signal that does not have background added in. This is presumably equivalent to an astronomer
        subtracting the background off of their images before scaling and subtracting the PSF subtraction source.
        """

        # Signal and noise in 2D. Get from extracted products
        s = self.extracted['detector_signal']
        n = self.extracted['detector_noise']
        self.detector_saturation = self.extracted['detector_saturation']

        # not reported, but needs to be available as a class attribute
        self.bg_pix = self.signal.rate_plus_bg - self.signal.rate

        # The noise is undefined if the pixel is fully saturated.
        # The engine currently returns noise=0 if the pixel has full saturation, which is
        # confusing since the noise is not 0, but rather undetermined or infinite. Setting
        # the noise to NaN ensures that the S/N of saturated pixels are NaNs.
        n[self.detector_saturation == 2] = np.nan

        self.detector_sn = s / n
        self.detector_signal = s + n * np.random.randn(n.shape[0], n.shape[1])

        self.r['2d']['detector'] = self.detector_signal
        self.r['2d']['snr'] = self.detector_sn
        self.r['2d']['saturation'] = self.detector_saturation
        self.r['2d']['ngroups_map'] = self.extracted['detector_ngroups']

    def _scalar(self):
        """
        Coronagraphy-specific scalar quantities.

        By design, engine inputs should be optional. That should include this contrast information, too.
        """

        Report._scalar(self)
        if 'contrast_separation' in self.input['strategy']:
            contrast_separation = self.input['strategy']['contrast_separation']
            self.r['scalar']['contrast_separation'] = contrast_separation
            self.r['scalar']['contrast_azimuth'] = self.input['strategy']['contrast_azimuth']
            self.r['scalar']['contrast'] = np.interp(contrast_separation, self.extracted['contrast_curve'][0],
                                                self.extracted['contrast_curve'][1])

    def _warnings(self):
        Report._warnings(self, sub_words={'': None, '1':'Science Scene', '2':'PSF Subtraction Source', '3':'Contrast Observation'})

class SpecApPhotReport(Report):
    """
    This report class is specifically for Aperture Spectral Extraction

    Parameters are read in through Report.__init__

    """
    def _scalar(self):
        Report._scalar(self)
        if isinstance(self.signal.current_instrument, HSTInstrument) and 'extraction_width' in self.extracted and 'extraction_height' in self.extracted:
            self.r['scalar']['extraction_width'] = self.extracted['extraction_width']
            self.r['scalar']['extraction_height'] = self.extracted['extraction_height']

class SpecSegmentedPhotReport(SpecApPhotReport):
    pass

class SpecScanApPhotReport(SpecApPhotReport):
    """
    This report class is specifically for Aperture Spectral Scan Extraction

    Parameters are read in through Report.__init__
    """

class MSAFullApPhotReport(SpecApPhotReport):
    """
    This report class is specifically for MSA Full Aperture Spectroscopy

    Parameters are read in through Report.__init__

    """
    pass


class MSAShutterApPhotReport(Report):
    """
    This report class is specifically for MSA Shutter Aperture Photometry 

    Parameters are read in through Report.__init__

    """
    pass


class MSAApPhotReport(Report):
    """
    This report class is specifically for MSA Shutter Aperture Photometry 
    
    Parameters are read in through Report.__init__

    """
    pass


class SOSSReport(SpecApPhotReport):
    """
    This report class is specifically for SOSS Parameters are read in through Report.__init__

    """
    pass


class IFUApPhotReport(Report):
    """
    This report class is specifically for IFU Aperture Photometry (and its derivatives) Parameters are read in
    through Report.__init__

    """

    def _information(self):
        """
        IFU Aperture Photometry-specific information.

        IFU modes are computed as spectroscopy, but reported as imaging (reconstructed data cubes)
        """
        Report._information(self)
        # this is used by the UI to decide how to make the plots. even though IFUs are spectroscopic,
        # they're plotted as 2D spatial images. they could be shown as either single planes from the
        # reconstructed cubes or some combination of multiple planes.
        self.r['information']['calc_type'] = "image"

    def _projection(self):
        """
        IFU Aperture Photometry-specific projection.

        IFU modes use extracted properties for their projection, and also require wavelength setting in a
        fashion similar to Aperture Spectral Extraction.
        """
        self.grid = self.signal.grid
        # for IFUs we want the reconstructed image plane that's build by the strategy
        self.pix_grid = self.extracted['plane_grid']
        self.wave_pix = self.extracted['wavelength']

    def _2d(self):
        """
        IFU Aperture Photometry-specific 2D products.

        For IFU modes, all of the 2d outputs of IFUs are linked to the 3D cubes and defined there.
        """
        pass


    def _3d(self):
        """
        IFU Aperture Photometry-specific 3D and 2D products.

        IFU modes report 3D data cubes, in addition to the input model cubes all other strategies report. The
        2D IFU data is taken from slices of these 3D cubes.
        """
        Report._3d(self)

        # adding a np.random.randn() instance here operating on self.extracted['detector_signal'], before it's run for
        # cube_sim, will get the same random numbers as the v1.2 and earlier version of Report.py
        # Signal and noise. Get from extracted products

        self.cube_signal, self.cube_noise, self.cube_saturation, self.cube_group_map, self.cube_plane_grid = self.extracted['reconstructed']
        self.cube_sim = self.cube_signal + self.cube_noise * np.random.randn(*self.cube_noise.shape)

        self.r['3d']['reconstructed'] = self.cube_sim
        self.r['3d']['reconstructed_signal'] = self.cube_signal
        self.r['3d']['reconstructed_noise'] = self.cube_noise
        self.r['3d']['reconstructed_saturation'] = self.cube_saturation
        self.r['3d']['reconstructed_ngroups_map'] = self.cube_group_map
        self.r['3d']['reconstructed_snr'] = self.cube_signal / self.cube_noise

        # use cube planes for IFUs, though this may be temporary. the actual detector image for an IFU
        # observation is a set of spectra, one for each IFU slice. Populate the detector and SNR images from the
        # planes of the reconstructed cubes, but collapse the saturation cube to show the most severe saturation
        # for each spatial pixel.
        # These 2D plots do not obey resels, they are the actual measured planes.

        self.r['2d']['detector'] = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.cube_sim, self.rw)
        self.r['2d']['snr'] = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.cube_signal, self.rw) / self.signal.the_detector.to_resels_scalar(self.wave_pix, self.cube_noise, self.rw)
        self.r['2d']['saturation'] = np.amax(self.cube_saturation, axis=0)
        self.r['2d']['ngroups_map'] = np.amin(self.cube_group_map, axis=0)
        # This is the background rate in each pixel without sources
        self.bg_pix = self.signal.rate_plus_bg - self.signal.rate

    def _3dfits(self):
        """
        IFU Aperture Photometry-specific 3D fits output.

        IFU modes report 3D data cubes, in addition to the input model cubes all other strategies report.
        """
        Report._3dfits(self)
        header = self.pix_grid.wcs_info()
        t = self.signal.spectral_detector_transform()
        header['ctype3'] = 'Wavelength'
        header['crpix3'] = 1
        header['crval3'] = t['wave_det_min'] - 0.5 * t['wave_det_step']
        header['cdelt3'] = t['wave_det_step']
        header['cunit3'] = 'um'
        header['cname3'] = 'Wavelength'
        for k in ['', '_signal', '_noise', '_snr', '_saturation']:
            key = 'reconstructed' + k
            # the reconstructed cubes are in the proper z, y, x order...
            # The Y axis must be flipped so the result looks the same in DS9 (etc) as
            #  it does in the client.
            o = fits.PrimaryHDU(self.r['3d'][key][:, ::-1, :])
            o.header.update(header)
            self.f['3d'][key] = o

    def _warnings(self):
        """
        Collect warnings for the IFU mode, and generate saturation warnings. This does not call to
        Report._warnings() because ALL of the saturation warnings for IFU modes are different from anywhere
        else.

        Parameters
        -----------
        sub_words : dict
            dict of strings to use when differentiating warning subreports
        """
        sub_words = {'': None, '1': 'Nod 1', '2': 'Nod 2'}
        self.r['warnings'] = self.warnings
        sat_obj = self.cube_saturation[self.wave_index]

        self.r['scalar']['n_partial_saturated'] = (sat_obj == 1).sum()
        self.r['scalar']['n_full_saturated'] = (sat_obj == 2).sum()

        self.r['scalar']['n_cube_partial_saturated'] = (self.cube_saturation == 1).sum()
        self.r['scalar']['n_cube_full_saturated'] = (self.cube_saturation == 2).sum()

        self._check_output_warnings(self.signal.current_instrument.warning_list_output, sub_word=sub_words[self.sub_str], sub_str=self.sub_str)

        # these properties were only needed for warning generation
        del self.r["scalar"]["n_partial_saturated"]
        del self.r["scalar"]["n_full_saturated"]
        del self.r['scalar']['n_cube_partial_saturated']
        del self.r['scalar']['n_cube_full_saturated']

class IFUNodInSceneReport(IFUApPhotReport):
    """
    This report class is specifically for IFUNodInScene (and its derivatives) 
    
    Parameters are read in through Report.__init__

    """
    pass

class IFUNodOffSceneReport(IFUNodInSceneReport):
    """
    This report class is specifically for IFUNodOffScene (and its derivatives) 
    
    Parameters are read in through Report.__init__

    """
    pass


class ImagingApPhotReport(Report):
    """
    This report class is specifically for Imaging Aperture Photometry (and its derivatives) 
    
    Parameters are read in through Report.__init__

    """
    pass


class ImagingScanApPhot(ImagingApPhotReport):
    """
    This report class is specifically for Imaging Scan Aperture Photometry (and its derivatives) 
    
    Parameters are read in through Report.__init__

    """
    pass

class TARectPhotReport(Report):
    """
    This report class is specifically for Target Acquisition Photocenter (and its derivatives) 
    
    Parameters are read in through Report.__init__

    """
    def _warnings(self):
        """
        TAPhot-specific warnings.

        Target Acquisition modes are uniquely concerned with saturation, because target acquisition is not
        possible if too many pixels on the detector are saturated. Likewise, the SNR must be higher than a
        certain value for target acquisition to succeed.
        """
        if hasattr(self.signal.current_instrument, "max_saturated_pixels"):
            self.r["scalar"]["n_total_saturated"] = self.extracted['saturation_products']['full'][0] + self.extracted['saturation_products']['partial'][0]
            self.r["scalar"]["max_saturated_pixels"] = self.signal.current_instrument.max_saturated_pixels
            self.r["scalar"]["min_snr_threshold"] = self.signal.current_instrument.min_snr_threshold
        Report._warnings(self)
        if hasattr(self.signal.current_instrument, "max_saturated_pixels"):
            del self.r["scalar"]["n_total_saturated"]
            del self.r["scalar"]["max_saturated_pixels"]
            del self.r["scalar"]["min_snr_threshold"]


class TACentroidReport(TARectPhotReport):
    """
    This report class is specifically for Target Acquisition Centroid Parameters are read in through
    Report.__init__

    """
    def _1d(self):
        """
        TACentroid-specific 1D products.

        Some strategies may not return a flux. TargetAcqCentroid is one example for now; it reports the
        centroid in place of the flux. The property is still called "extracted_flux" in the curves, but we
        rename them here to avoid confusion. The "extracted_noise" is still the standard deviation of the
        extracted product.
        """
        Report._1d(self)

        self.r['scalar']['extracted_centroid'] = float(self.extracted['extracted_flux'][self.wave_index])
        self.r['scalar']['extracted_flux'] = np.nan


class TASpecReport(ImagingApPhotReport):
    """
    This report class is specifically for Spectral Target Acqusition (and its derivatives) 
    
    Parameters are read in through Report.__init__

    """
    def _1d(self):
        Report._1d(self)

        if self.sub_str == "":
            if "acq_mode" in self.input["strategy"] and self.input["strategy"]["acq_mode"] == "acq_peakxd":
                if "stripe" in self.input["strategy"]:
                    report_mapping = {"a": 0, "b": 1, "c": 2}
                    subreport = report_mapping[self.input["strategy"]["stripe"]]
                    self.r['scalar']['sn_per_stripe'] = {}
                    self.r['scalar']['time_per_stripe'] = {}
                    self.r['scalar']['reference_wavelength_stripe'] = {}
                    # flip the mapping list so we can generate stripe names from report items. This should be more robust against a different number of sub_reports.
                    mapping_report = {v: k for k, v in report_mapping.items()}
                    for subreport in range(len(self.extracts)):
                        signal = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracts[subreport]['extracted_flux'], self.rw)
                        noise = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracts[subreport]['extracted_noise'], self.rw)
                        self.r['scalar'][f'sn_per_stripe'][mapping_report[subreport]] =  signal/noise
                        self.r['scalar'][f'time_per_stripe'][mapping_report[subreport]] = self.extracts[subreport]['all_dithers_time']
                        self.r['scalar'][f'reference_wavelength_stripe'][mapping_report[subreport]] = self.extracts[subreport]['wavelength']
            elif self.input["configuration"]["instrument"]["mode"] == "fuv_specacq":
                mapping_report = {0: "fuv-b", 1: "fuv-a", 2: "combined"}
                if len(self.extracts) == 2: #in the case of only two subreports (G140L), the first is fuv-a
                    mapping_report = {0: "fuv-a" , 1: "combined"}
                self.r['scalar']['sn_per_stripe'] = {}
                self.r['scalar']['time_per_stripe'] = {}
                self.r['scalar']['reference_wavelength_stripe'] = {}
                for subreport in range(len(self.extracts)):
                    signal = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracts[subreport]['extracted_flux'], self.rw)
                    noise = self.signal.the_detector.to_resels_scalar(self.wave_pix, self.extracts[subreport]['extracted_noise'], self.rw)
                    self.r['scalar'][f'sn_per_stripe'][mapping_report[subreport]] =  signal/noise
                    self.r['scalar'][f'time_per_stripe'][mapping_report[subreport]] = self.extracts[subreport]['all_dithers_time']
                    self.r['scalar'][f'reference_wavelength_stripe'][mapping_report[subreport]] = self.extracts[subreport]['wavelength']

    def _bop(self):
        """
        Compute HST's Bright Object Protection (Health and Safety) quantities
        """
        if self.sub_str == "":
            if "acq_mode" in self.input["strategy"] and self.input["strategy"]["acq_mode"] != "acq_peakxd" or "stripe" not in self.input["strategy"]:
                Report._bop(self)
            else:
                if "stripe" in self.input["strategy"]:
                    report_mapping = {0: "nuv-a", 1: "nuv-b", 2: "nuv-c"} # this maps stripe names to the subreport we should read them from

                    brightest_pixel_rate = {}
                    detector_total_rate = {}
                    bright_pixel_rates = []
                    for stripe in range(len(self.r["sub_reports"])):
                        bright_pixel_rates.append(self.r["sub_reports"][stripe]["bop"]["brightest_pixel_rate"][report_mapping[stripe]][0])
                        brightest_pixel_rate[stripe] = self.r["sub_reports"][stripe]["bop"]["brightest_pixel_rate"][report_mapping[stripe]]
                        detector_total_rate[stripe] = self.r["sub_reports"][stripe]["bop"]["detector_total_rate"][report_mapping[stripe]]
                    self.r['bop']['brightest_pixel_rate'] = brightest_pixel_rate
                    self.r['bop']['detector_total_rate'] = detector_total_rate

                    # TODO  -  JETC-2484
                    # This is a temporary bail-out code that prevents bop quantities to be calculated
                    # when the flux cube contains nans. Otherwise, the code downstream that expects
                    # bop values, crashes. To bem removed once the ticket above gets implemented.
                    if len(self.signal.brightest_pixel_rate.values()) == 0:
                        return

                    # BOP brightest pixel rate and detector total rate are provided for more than one segments
                    # find the highest per-segment value and add BOP warning based on it

                    # max brightest pixel rate from the dict with tuples for rate and wavelength at which it occurs for every chip
                    time = self.signal.the_detector.exposure_spec.exposure_time # for a single exposure, taking nsplit into account
                    max_stripe = np.argmax(bright_pixel_rates)
                    time = self.r["sub_reports"][max_stripe]["scalar"]["exposure_time"]

                    self.r['bop']['max_bright_pixel_rate'] = np.max(bright_pixel_rates)
                    self.r['bop']['max_bright_pixel_count'] = np.max(bright_pixel_rates) * time
                    self.r['bop']['max_total_detector_rate'] = max(self.r['bop']['detector_total_rate'].values())
        else:
            Report._bop(self)


class TAPhotReport(TARectPhotReport):
    pass

class TABoxPhotReport(TARectPhotReport):
    pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
io_utils - commonly used I/O routines for JSON, FITS, and possibly other formats.
"""

import os
import json
import errno
import numpy as np
import astropy.io.fits as fits
import astropy.units as u

from .custom_exceptions import EngineInputError, DataError
from .constants import PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS, UNIT_MAP


class NumPyArangeEncoder(json.JSONEncoder):

    """
    custom encoder to handle numpy arrays that might show up in inputs and outputs:
    http://stackoverflow.com/questions/11561932/why-does-json-dumpslistnp-arange5-fail-while-json-dumpsnp-arange5-tolis
    Parameters
    ----------
    obj: python object
        currently only supports translating np.ndarray and np.float32 objects
    Returns
    -------
    e: json.JSONEncoder object
        JSONEncoder that now understands the custom data types
    """

    def default(self, obj):
        # unpack Quantities first, then process them as needed
        if isinstance(obj, u.quantity.Quantity):
            obj = obj.value
        # turn ndarrays into lists
        if isinstance(obj, np.ndarray):
            return obj.tolist()  # or map(int, obj)
        # numpy.float32's started showing up when refactoring refdata FITS files
        # catch them and make them into normal float()'s
        if isinstance(obj, (np.float32, np.float64)):
            return float(obj)
        if isinstance(obj, (np.int32, np.int64)):
            return int(obj)
        # work the mapping backwards to get strings we recognize
        if isinstance(obj, (u.core.PrefixUnit, u.core.Unit)):
            return [key for (key, value) in UNIT_MAP.items() if value == obj][0]
        e = json.JSONEncoder.default(self, obj)
        return e

def get_full_refwave_dict(data_dir, telescope, inst_list, modename_list):
    """
    Traverse the pandeia_data tree for the instruments and modenames given
    and find all refwave info (min, max, midpoint) that there is and return it in
    a tree of dicts.

    wavelength ranges are stored in "range", except for when there are slit-dependent 
    wavelength gaps (currently, only JWST NIRSpec), where it's read out of "slit_config"

    Parameters
    ----------
    data_dir: str
        Path to pandeia_data dir (e.g. $pandeia_refdata value)
    telescope: str
        E.g. 'jwst'
    inst_list: list of str
        instrument names for data desired, e.g. ['miri','nircam','niriss','nirspec'],
    modename_list: list of str
        mode names desired for all instruments listed above, e.g. ['imaging','mrs','soss','wfgrism']
    """
    def add_midpoint_key(adict):
       """ convenience function used a few times in here. handle one- or many-level dicts """
       if not isinstance(adict, dict):
           raise ValueError('Unexpected non-dict input: %s' % adict)
       if 'wmin' in adict and 'wmax' in adict:
           mid = (adict['wmin'] + adict['wmax'])/2.0
           mid = (int(1000*mid))/1000. # keep only to 3rd decimal place
           adict['wmidpoint'] = mid
       for key in adict: # now recurse to catch any wmin/wmax sub-dicts
           if isinstance(adict[key], dict):
               add_midpoint_key(adict[key])

    retval = {}
    for inst in inst_list:
        this_inst = {}
        inst_cfg_fname = data_dir+os.sep+telescope+os.sep+inst+os.sep+'config.json'
        assert os.path.exists(inst_cfg_fname), "Error: cfg file not found: "+inst_cfg_fname
        inst_cfg = read_json(inst_cfg_fname, raise_except=True)
        all_inst_modes = inst_cfg['modes']
        mode_cfg = inst_cfg['mode_config']
        range_cfg = inst_cfg['range']
        assert sorted(all_inst_modes) == sorted(mode_cfg.keys()) or (inst == 'nirspec' and 1 + len(all_inst_modes) == len(mode_cfg)), "Error: modes list mismatch for: %s \n   all-inst-modes = %s, but mode_cfg has: %s" % (inst, sorted(all_inst_modes), sorted(mode_cfg))
        requested_modes = [m for m in modename_list if m in all_inst_modes]
        for mode in requested_modes:
            this_mode = {}
            inst_mode_apers = mode_cfg[mode]['apertures']
            for aper in inst_mode_apers:
                this_aper = {}
                this_slit = inst_cfg['aperture_config'][aper].get('slit', None)
                if mode != 'msa' and 'slit_config' in inst_cfg and this_slit in inst_cfg['slit_config'] and 'gap' in inst_cfg['slit_config'][this_slit]: # check for the gaps dict
                    # handle dispersed situations with gaps differently, e.g. nirspec (but not yet msa/mos, #4645)
                    full_slit_wave_tree = {}
                    # since these wave ranges can (in some modes) vary by another param (e.g. subarray),
                    # we collect all the sub_key-specific trees (one will be called 'default')
                    for sub_key in inst_cfg['slit_config'][this_slit]['gap']:
                        wave_tree = inst_cfg['slit_config'][this_slit]['gap'][sub_key]
                        # Massage (flatten & simplify) the dict tree to make the reading logic simpler.
                        # For now ignore gap_* keys since our callers do not need them. (can add later)
                        flat_sub_tree = {}
                        if sub_key != 'key_format':
                            for grism_key in wave_tree:
                                # flatten these - make no use of the sub-tree ability here employed in pandeia_data
                                if 'wave_min' in wave_tree[grism_key]: # grism only, no filters defined (so covers all poss. filter choices)
                                    flat_sub_tree[grism_key+',*'] = {'wmin': wave_tree[grism_key]['wave_min'], 'wmax': wave_tree[grism_key]['wave_max']}
                                else: # possible multiple filters for same grism, e.g. like 'g140h,f070lp'
                                    for filtkey in wave_tree[grism_key]:
                                        assert isinstance(wave_tree[grism_key][filtkey], dict), 'Unexpected layout in wave_tree for %s.%s' % (grism_key, filtkey)
                                        flat_sub_tree[grism_key+','+filtkey] = \
                                            {'wmin':  wave_tree[grism_key][filtkey]['wave_min'], 'wmax': wave_tree[grism_key][filtkey]['wave_max']}
                                for k in flat_sub_tree:
                                    add_midpoint_key(flat_sub_tree[k])
                        else:
                            flat_sub_tree = wave_tree # is just a scalar string
                        full_slit_wave_tree[sub_key] = flat_sub_tree
                    this_aper = full_slit_wave_tree
                elif aper in range_cfg:
                    # we have matching range data !
                    for item in range_cfg[aper]:
                        if isinstance(range_cfg[aper][item], dict):
                            # Handle whether this is a simple wmin/wmax dict, or whether this has yet one 
                            # more level to it (with wmin/wmax dicts below). e.g. soss
                            akey = None
                            if len(range_cfg[aper][item]):
                                akey = list(range_cfg[aper][item].keys())[0] # need only check layout of 1 key subdict
                            if ('wmin' in range_cfg[aper][item] and 'wmax' in range_cfg[aper][item]) or \
                               (akey in range_cfg[aper][item] and isinstance(range_cfg[aper][item][akey], dict) and \
                                'wmin' in range_cfg[aper][item][akey] and 'wmax' in range_cfg[aper][item][akey]):
                                this_item = range_cfg[aper][item].copy() # use a copy for below
                                add_midpoint_key(this_item)
                                this_aper[item] = this_item
                # done with aper
                if len(this_aper) > 0:
                    this_mode[aper] = this_aper
            # done with mode
            if len(this_mode) > 0:
                this_inst[mode] = this_mode
        # done with inst
        if len(this_inst) > 0:
            retval[inst] = this_inst
    return retval


def read_json(filename, raise_except=False, **kwargs):
    """
    read in a JSON format file.  return None if the file is not there.

    Parameters
    ----------
    filename: string
        name of the JSON file to read
    except: bool
        if true, raise exception if file is missing. if false, return empty dict.

    Returns
    -------
    d: python object
        data from JSON file decoded into python object
    """
    try:
        with open(filename, 'r') as f:
            json_data = json.load(f, **kwargs)
    except IOError as e:
        if e.errno == errno.ENOENT and raise_except is False:  # No such file
            json_data = {}
        else:
            msg = "Missing JSON file: %s" % filename
            raise EngineInputError(value=msg)
    d = json_data
    return d


def write_json(data, filename, **kwargs):
    """
    write python object into a JSON format file

    Parameters
    ----------
    data: python object
        python object to encode and write to JSON
    filename: string
        name of file to write JSON-encoded data to
    """
    with open(filename, 'w') as f:
        json.dump(data, f, indent=4, cls=NumPyArangeEncoder, separators=(',', ': '), **kwargs)
        f.write('\n')


def append_json(obj, filename):
    """
    read data structure from JSON, append it to a python object, and return the result.
    currently only supported for lists and dicts

    Parameters
    ----------
    obj: python object
        object to be appended to
    filename: string
        filename containing JSON data to be appended

    Returns
    -------
    obj: python object
        input python object now updated with appended data
    """
    json_data = read_json(filename)
    if isinstance(obj, list):
        if isinstance(json_data, list):
            for k in json_data:
                obj.append(k)
        else:
            obj.append(json_data)
    elif isinstance(obj, dict):
        if isinstance(json_data, dict):
            obj.update(json_data)
    else:
        raise ValueError("Can only append to list or dict.")
    return obj


def convert_nd2list_in_place(obj):
    """
    Traverse a dict (with possible nested dicts) finding any numpy
    ndarray instances and converting them to Python lists (or lists of lists),
    but do so in place without making a second copy of the data structure.

    Returns nothing; modifies the dict passed in.

    Parameters
    ----------
    obj: python dict
        data structure to be traversed

    Returns
    -------
    Nothing
    """
    if isinstance(obj, dict):
        thekeys = list(obj.keys()) # intentional copy of keys
        for key in thekeys:
            if isinstance(obj[key], np.ndarray):
                as_list = obj[key].tolist()
                del obj[key]
                obj[key] = as_list
            elif isinstance(obj[key], list) and len(obj[key]) > 0 and isinstance(obj[key][0], np.ndarray):
                # here we have a list of arrays, so convert them all
                as_list_list = []
                for item in obj[key]:
                    as_list_list.append(item.tolist())
                del obj[key]
                obj[key] = as_list_list
            elif isinstance(obj[key], dict):
                # recurse !
                convert_nd2list_in_place(obj[key])
            else:
                pass # leave this key of dict untouched
    else:
        pass


def ref_data_interp(filename, wave, colname=None):
    """
    Read reference data from a FITS file and interpolate it to a provided wavelength array

    Parameters
    ----------
    filename: str
        Filename of reference data
    wave: numpy.ndarray
        Wavelength vector that the reference data will be interpolated onto
    colname: str
        Name of column within the reference file to read and interpolate

    Returns
    -------
    interp_col: numpy.ndarray
        Vector containing reference data interpolated onto wave
    """
    if colname is None:
        raise EngineInputError(value="Must specify name of column to read from reference file.")
    try:
        data = fits.getdata(filename)
    except IOError as e:
        error_msg = "Error reading reference file: " + filename
        raise DataError(value=error_msg)
    if np.any(np.diff(data['wavelength']) < 0):
        indices = np.where(np.diff(data['wavelength']) < 0)[0]
        error_msg = "Wavelengths must be increasing in reference file: %s\n" % (filename)
        error_msg += "Out-of-order indices: %s" % repr(indices)
        raise DataError(value=error_msg)
    try:
        columns = set(k.name.lower() for k in data.columns)
        if colname.lower() not in columns:
            msg = "Column %s not found in %s" % (colname, filename)
            raise DataError(value=msg)
        interp_col = np.interp(wave, data['wavelength'], data[colname])
    except Exception as e:
        error_msg = "Error interpolating reference file: %s : %s" % (filename, type(e))
        raise DataError(value=error_msg)
    return interp_col


def ref_data_column(filename, colname=None, error_msg="Error loading reference file."):
    """
    Read reference data from a FITS file and provide it directly with no interpolation

    Parameters
    ----------
    filename: str
        Filename of reference data
    colname: str
        Name of column within the reference file to read and return
    error_msg: str
        Custom error message to produce in case of an error

    Returns
    -------
    col: numpy.ndarray
        Vector containing requested reference data
    """
    if colname is None:
        raise EngineInputError(value="Must specify name of column to read from reference file.")
    try:
        data = fits.getdata(filename)
    except IOError as e:
        raise DataError(value="I/O " + error_msg)
    try:
        col = data[colname]
    except KeyError as e:
        raise DataError(value="Column not found in reference file: %s" % colname)
    return col


def mkdir_p(path):
    """
    Implement 'mkdir -p' functionality with pure python

    Parameters
    ----------
    path: valid path specification
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise




import astropy.units as u
import stsynphot.spparser as sp

from pandeia.engine.constants import PANDEIA_WAVEUNITS
from pandeia.engine.custom_exceptions import SynphotError

class SkySpectrum(object):

    def __init__(self, synphotExpr):
        self.spectrum = self._make_continuum(synphotExpr)

    def _make_continuum(self, synphotExpr):
        if isinstance(synphotExpr, str):
            try:
                result = sp.parse_spec(synphotExpr)
            except Exception as e:
                message = self.name + " " + str(e).lower()
                raise SynphotError(value=message)
        else:
            raise TypeError("Cannot create spectrum from %s"%type(synphotExpr))
        return result

    @property
    def wave_in_microns(self):
        # convert from angstroms to microns
        return self.spectrum.waveset.to_value(PANDEIA_WAVEUNITS)

    @property
    def flux_in_MJy_per_sr(self):
        # convert flux from flam/arcsec^2 to MJy/sr
        # step 1: convert from flam to MJy (megaJanskies)
        flux = self.spectrum(self.spectrum.waveset, flux_unit=u.MJy).to_value(u.MJy)
        # step 2: convert from arcsec^2 to steradians
        flux = flux / (u.arcsec * u.arcsec).to(u.sr)
        return flux



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Module that implements a single calculation API for the Pandeia Engine
"""

from .etc3D import calculate_exposure_time, calculate_sn
from . import debug_utils

def perform_calculation(calc_input, dict_report=True, webapp=True, validate=False):
    """
    Function to perform a single ETC calculation

    Parameters
    ----------
    calc_input: dict
        Dictionary containing the information required to perform the calculation.
    dict_report: Boolean (default: True)
        If True, return a dict in engine output API format. Otherwise return
        a report.Report instance.
    webapp: Boolean (default: True)
        Toggle strict API checking for webapp
    validate: Boolean (default: False)
        Toggle whether additional intermediate products are saved internally
    """

    if True:   # you can change it to False
        if 'fake_exception' in calc_input:
            perform_fake_exceptions(calc_input)

    debug_utils.init(calc_input)

    # run the calculation....
    snr = calc_input['configuration']['detector'].get('calculate_snr', True)
    if snr:
        report = calculate_sn(calc_input, webapp=webapp, validate=validate)
    else:
        report = calculate_exposure_time(calc_input, webapp=webapp, validate=validate)

    if dict_report:
        return report.as_dict()
    else:
        return report


def perform_fake_exceptions(calc_input):
    i = calc_input['fake_exception']
    if 'pandeia' in i:
        from . import custom_exceptions
        raise custom_exceptions.PandeiaException("fake pandeia exception for testing")
    if 'exception' in i:
        raise Exception("fake abnormal exception for testing")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import operator as op
import numpy as np
import astropy.constants as cs
import scipy.integrate as integ
import scipy.interpolate as interp
import scipy.signal
from astropy.io import fits
from astropy.table import Table
import astropy.units as u
import synphot as syn

from . import config as cf
from .normalization import NormalizationFactory
from .profile import ProfileFactory
from .extinction import ExtinctionFactory
from .sed import SEDFactory
from .coords import Grid
from .utils import spectrum_resample
from .custom_exceptions import EngineInputError, WavesetMismatch, DataError, RangeError, DataConfigurationError
from .pandeia_warnings import astrospectrum_warning_messages as warning_messages
from .constants import PANDEIA_RTOL, PANDEIA_ATOL, CONVERSION_CONSTANT, SPECTRAL_LINE_WINDOW, SPECTRAL_LINE_SUBSAMPLE, SPECTRAL_MAX_SAMPLES, SPECTRAL_MAX_CHUNK, SCENE_FOV_PIXBUFFER, SCENE_MAX_PIXELVAL, PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS

default_refdata_directory = cf.default_refdata_directory
# pyfftw.interfaces.cache.enable()

class ModelSceneCube(object):

    """
    This generates a model intensity cube of a scene. It takes the analytic descriptions
    of the components of the scene and samples them into a Grid.

    Here we put wavelength as the 3rd index to enable broadcasting. The cube is a stack
    of images, each at a different wavelength. If we wish to add or multiply a wavelength
    vector (e.g. throughput) with the cube, the cube needs to be in this index order for
    it to work efficiently. FITS, however, expects an order of (wave, y, x) so a transpose
    will be required to convert to that format. Other formats may have similar requirements.

    Parameters
    ----------
    source_spectra: list of pandeia.engine.astro_spectrum.AstroSpectrum instances
        Source spectra to add to the model cube
    grid: pandeia.engine.coord.Grid instance
        Grid describing the 2D spatial coordinates for each plane of the cube
    psf_upsamp: int
        How oversampled the grid is. Needed to generate a realistic brightest pixel rate

    Methods
    -------
    add_source:  add an AstroSpectrum to the cube
    export_to_fits: export model cube to a FITS file
    """

    def __init__(self, source_spectra, grid, psf_upsamp):
        self.grid = grid
        self.x = grid.x
        self.y = grid.y
        self.xsamp = grid.xsamp
        self.ysamp = grid.ysamp
        self.nx = grid.nx
        self.ny = grid.ny
        self.psf_upsamp = psf_upsamp
        self.wave = source_spectra[0].wave
        self.nw = self.wave.shape[0]
        self.central_profiles_list = []
        # now go through the source spectra and make sure they've been merged
        # onto the same wavelength set
        for s in source_spectra:
            if not np.array_equal(self.wave, s.wave):
                message = "Model cube input spectra must be sampled at the same wavelengths."
                raise EngineInputError(value=message)
        # stick wavelength as the 3rd index of the cube to enable broadcasting
        self.int = np.zeros(self.x.shape + (self.wave.size, ), dtype=np.float32)

        for source_spectrum in source_spectra:
            self.add_source(source_spectrum)

    def add_source(self, spectrum):
        """
        Add a source to the model cube.

        Parameters
        ----------
        spectrum: pandeia.engine.astro_spectrum.AstroSpectrum instance
            Spectrum object containing spatial and spectral information for the source to be added
        """
        src = spectrum.src
        src.grid = self.grid
        self.profile = ProfileFactory(self.psf_upsamp, config=src)
        plane = self.profile.normalized()
        central_pixel, extended = self.profile.brightest()
        self.central_profiles_list.append(central_pixel)

        # This broadcasting step can really spike the memory for large fields (like SOSS).
        # So we split up the task if there are a lot of wavelength planes at the cost of a few seconds of run time.
        # Potentially this could be done more elegantly
        if self.nw>SPECTRAL_MAX_CHUNK:
            self.int[:,:,0:SPECTRAL_MAX_CHUNK] += plane.reshape(plane.shape + (1,)) * spectrum.flux[0:SPECTRAL_MAX_CHUNK]
            self.int[:,:,SPECTRAL_MAX_CHUNK:] += plane.reshape(plane.shape + (1,)) * spectrum.flux[SPECTRAL_MAX_CHUNK:]
        else:
            self.int += plane.reshape(plane.shape + (1, )) * spectrum.flux

    def export_to_fits(self, fitsfile='ModelSceneCube.fits'):
        """
        Write model cube to a FITS file
        """
        header = self.grid.wcs_info()
        t1 = fits.PrimaryHDU(np.moveaxis(self.int, [0, 1, 2], [1, 2, 0]))
        c1 = fits.Column(name='wavelength', array=self.wave, format='D')
        t2 = fits.BinTableHDU.from_columns([c1])
        t1.header.update(header)
        tbhdu = fits.HDUList([t1,t2])
        tbhdu.writeto(fitsfile, overwrite=True)
        tbhdu.close()


class AstroSpectrum(object):

    """
    Class that contains a 1D wavelength grid and methods to calculate a single point spectrum,
    for instance from a point source.

    Parameters
    ----------
    src: pandeia.engine.source.Source
        Source instance containing parameters relevant for the definition of the astro spectrum.
    webapp: bool
        Determines whether extra API checks are done

    Attributes
    ----------
    src: pandeia.engine.source.Source instance
        The input source containing parameters relevant for the definition of the astro spectrum.
    line_definition: list of pandeia.engine.source.Line instances
        The lines contained in the spectrum of self.src
    normalization: instance of pandeia.engine.normalization.Normalization subclass
        Normalization information defining the brightness of self.src
    sed: instance of pandeia.engine.sed.SED subclass
        Spectral energy distribution information for self.src (excluding lines)
    wave: 1D np.ndarray
        The wavelength vector of the spectrum
    flux: 1D np.ndarray
        The flux vector of the spectrum
    nw: int
           Number of wavelength points

    Methods
    -------
    export_to_fits_table()
        Write self.wave and self.flux to binary FITS table
    add_spectral_lines()
        Add spectral lines to the spectrum
    resample()
        Resample spectrum to a new set of wavelengths via interpolation
    trim()
        Trim spectrum to a specified wavelength range

    Notes
    -----

    """

    def __init__(self, src, webapp=False, **kwargs):

        self.src = src
        self.warnings = {}
        # get the spectrum information from the source
        self.line_definitions = sorted(self.src.spectrum['lines'], key=op.itemgetter("center"))
        self.normalization = NormalizationFactory(config=self.src.spectrum['normalization'], webapp=webapp)
        self.extinction = ExtinctionFactory(config=self.src.spectrum['extinction'], webapp=webapp)

        # if redshift is provided (it should be if config file is up-to-date) use it, else assume it's 0.0
        if 'redshift' in self.src.spectrum:
            z = self.src.spectrum['redshift']
        else:
            z = 0.0

        if z <= -1.0:
            msg = "Specified source redshift must be > -1.0."
            raise EngineInputError(value=msg)

        self.sed = SEDFactory(config=self.src.spectrum['sed'], webapp=webapp, z=z)

        self.wave, self.flux = (1.0 + z) * self.sed.wave, self.sed.flux
        self.nw = self.wave.size

        # Optional capability: the keyword has to exist and be set to false to have the non-default behavior
        if "extinction_first" in self.src.spectrum and self.src.spectrum["extinction_first"] == False:
            # apply normalization before extinction
            self.wave, self.flux = self.normalization.normalize(self.wave, self.flux)

            # apply extinction after redshift
            self.wave, self.flux = self.extinction.extinction(self.wave, self.flux)
        else:
            # apply extinction after redshift
            # the normalization is passed in so the bandpass can be obtained through its helper functions
            self.wave, self.flux = self.extinction.extinction(self.wave, self.flux)

            # apply normalization after extinction
            self.wave, self.flux = self.normalization.normalize(self.wave, self.flux)

        # update warnings...
        self.warnings.update(self.normalization.warnings)
        self.warnings.update(self.sed.warnings)

        # add lines after normalizing continuum since line strengths are currently specified
        # using physical units. future support for specifying line strengths as equivalent widths
        # would allow lines to be added before normalization.
        if len(self.line_definitions) > 0:
            self.add_spectral_lines()

        # Astropy Synphot usage ends here, so convert back to arrays
        self.wave = self.wave.to_value(PANDEIA_WAVEUNITS)
        self.flux = self.flux.to_value(PANDEIA_FLUXUNITS)

    def export_to_fits_table(self, fitsfile="source_spectrum.fits"):
        c1 = fits.Column(array=self.wave, format="D", name="wavelength", unit="micron")
        c2 = fits.Column(array=self.flux, format="D", name="flux", unit="mJy")
        tbhdu = fits.BinTableHDU.from_columns([c1, c2])
        tbhdu.writeto(fitsfile, overwrite=True)

    def add_spectral_lines(self):
        """
        This method adds lines to the spectrum by either adding the line flux (for emission lines) or calculating
        an absorption line using the optical depth, F = Fcont * exp(-tau). The sampling of the spectrum is checked for each line
        and new samples added as necessary to assure adequate sampling for each line. The sampling of the underlying spectrum is
        doubled to avoid any Nyquist smoothing effects.
        """
        # make a nyquist sampled version of the underlying wavelength set to mitigate the effects of resampling.
        # synphot is used for the resampling so flux is properly preserved in the process.
        halves = self.wave[0:-1] + 0.5 * np.diff(self.wave)

        # known issue with astropy: numpy's append function results in a Nonetype unit: http://docs.astropy.org/en/latest/known_issues.html#quantities-lose-their-units-with-some-operations
        nywave = np.append(self.wave.value, halves.value) * self.wave.unit
        nywave.sort()

        lines = []

        # loop through line definitions to build waveset properly sampled for all of them
        for line_definition in self.line_definitions:
            spectral_line = SpectralLine(line_definition)
            lines.append(spectral_line)
            # update wavelength set so that it optimally samples the line's wavelength region.
            nywave = spectral_line.line_waveset(nywave)

        # resample underlying spectrum to new wavelength set.
        # this method will update self.wave, self.flux, and self.nw accordingly.
        nyflux = np.nan_to_num(spectrum_resample(self.flux.value, self.wave.value, nywave.value))
        self.wave = nywave
        self.flux = nyflux << PANDEIA_FLUXUNITS
        self.nw = nywave.size

        # now apply the lines to the spectrum
        for spectral_line in lines:
            if spectral_line.emission_or_absorption == 'emission':
                self.flux = self.flux + spectral_line.flux(self.wave)
            elif spectral_line.emission_or_absorption == 'absorption':
                self.flux = self.flux * np.exp(-spectral_line.optical_depth(self.wave))
            else:
                raise EngineInputError(value="Invalid spectral line type: %s" % spectral_line.emission_or_absorption)

    def trim(self, wmin, wmax):
        """
        Trim spectrum to wavelength range specified by wmin and wmax

        Parameters
        ----------
        wmin: float
            Minimum wavelength (microns)
        wmax: float
            Maximum wavelength (microns)
        """
        valid_subs = np.where((self.wave >= wmin) & (self.wave <= wmax))
        if valid_subs is None or len(valid_subs[0]) == 0:
            # corner case: We have a spectrum that covers the wavelength range 
            # of the instrument setup, but it's so sparsely specified that no 
            # points fall within the range. (HETC-454)
            if (self.wave[0] <= wmin) and (self.wave[-1] >= wmax):
                wave = np.asarray([wmin,wmax])
                valid_subs = np.where((wave >= wmin) & (wave <= wmax))
                # 1D (linear) interpolation is all synphot does; I see no reason 
                # to create a spectrum here just to resample it
                int_spec = interp.interp1d(self.wave,self.flux)
                self.flux = int_spec(wave)
                self.wave = wave
            else:
                msg = "Spectrum for source does not have any overlap with instrument wavelength range."
                raise WavesetMismatch(value=msg)

        trim_wave = []
        # make sure trimmed wavelengths include wmin and wmax
        if not np.isclose(self.wave[valid_subs][0], wmin, atol=PANDEIA_ATOL, rtol=PANDEIA_RTOL):
            trim_wave.append(wmin)

        trim_wave.extend(self.wave[valid_subs])

        if not np.isclose(self.wave[valid_subs][-1], wmax, atol=PANDEIA_ATOL, rtol=PANDEIA_RTOL):
            trim_wave.append(wmax)

        trim_wave = np.array(trim_wave)

        # our spectra are in mJy, which is a flux DENSITY, so interpolation is appropriate for 
        # trimming (JETC-2812)
        int_spec = interp.interp1d(self.wave, self.flux, bounds_error=False, fill_value=np.nan)
        trim_flux = int_spec(trim_wave)

        self.wave = trim_wave
        self.flux = trim_flux
        self.nw = self.wave.size


class SpectralLine(object):

    """
    A spectral line - that is the information and methods needed to add a single line to a spectrum.
    The line could also form a component of a composite line (for instance a P Cygni line consisting
    of an emission line AND an absorption line). In this case, one would create two different instances
    of SpectralLine, with relevant profile parameters, and add them both to the spectrum. The class
    includes methods to create a wavelength window with finer sampling than the rest of the spectrum.
    The single line is modeled as a gaussian profile (only option right now, could be more later)
    with a FWHM, a central wavelength and a strength. The line can be either in emission or
    absorption. In the case of an emission line, the strength is an integrated line intensity in
    cgs units, and in the case of an absorption line the strength is a peak optical depth.

    Arguments
    definition - A dictionary containing the line parameters (center, width, strength, profile and emission_or_absorption).
    """

    def __init__(self, definition):
        """
        This needs to be updated to match configuration patterns used elsewhere in engine.  See #1904 for details.
        """
        self.center = definition['center'] * PANDEIA_WAVEUNITS
        self.width = definition['width'] * (u.km * u.second**-1)
        self.strength = definition['strength'] * u.mJy
        self.profile = definition['profile']
        self.emission_or_absorption = definition['emission_or_absorption']
        self.name = definition['name'] if 'name' in definition else ''

        # In wavelength units
        self.wave_width = self.width / cs.c * self.center

        # Run the sanity checks to make sure things are kosher.
        self._sanity_checks()

    def _sanity_checks(self):

        # Check the center, width and strength
        # Based on Issue #2263#issuecomment-263374621
        if self.center <= 0 or self.width <= 0 or self.strength < 0:
            msg = 'Invalid line configuration'
            if self.name:
                msg += ' for "' + self.name + '"'
            msg += ': Center must be > 0, width must be > 0 and strength must be >= 0.'
            raise EngineInputError(msg)

    def line_waveset(self, wave, window_factor=SPECTRAL_LINE_WINDOW):
        """
        Create set of wavelengths to optimally sample the spectral line.

        Arguments
        ---------
        wave: np.ndarray
            Wavelength set of spectrum to which line will be added
        window_factor: int
            Configure the size of the wavelength window over which wavelength samples are checked and, if necessary,
            created. Window size is 2 * window_factor * SPECTRAL_LINE_SUBSAMPLE.

        Returns
        -------
        lwave: np.ndarray
            Array of wavelengths
        """
        nsamp = window_factor * 2 * SPECTRAL_LINE_SUBSAMPLE  # Nyquist plus some oversampling
        wmin = np.max(self.center - self.wave_width * window_factor, 0)
        wmax = self.center + self.wave_width * window_factor

        # check input wavelength set over the line window to see if we need supplement with some more samples
        lsubs = (wave > wmin) & (wave < wmax)
        if len(wave[lsubs]) > nsamp:
            # if we're already sampled sufficiently over the line region, return wave unmodified
            # though just the units, to match the behavior of merge_wavelengths
            lwave = wave.value
        else:
            # otherwise create a finely sampled set over the line, merge with the input, and return result
            # NOTE (JETC-1668): lines (and therefore the additional line wavelengths) are specified in 
            # PANDEIA_WAVEUNITS (microns) but the spectrum is not yet guaranteed to be (and is most likely 
            # in Angstroms). We need to be absolutely sure the wavelengths are in the same units 
            # before merging, otherwise resampling the line flux might fail. This caused a bug present in 
            # v1.5.2 and fixed in v1.6
            lwaveset = np.linspace(wmin, wmax, int(nsamp))
            lwave = syn.utils.merge_wavelengths(wave.value, lwaveset.to_value(wave.unit))

        return lwave*wave.unit

    def flux(self, wave):
        """
        Calculate emission line flux over wavelength set, wave.

        Arguments
        ---------
        wave: np.ndarray
            Wavelengths over which to calculate line emission

        Returns
        -------
        flux: np.ndarray
            Line flux at wave
        """
        sigma = self.wave_width / 2.3548

        # normalized to a peak flux density of 1 mJy - note that flux will be unitless
        flux = np.exp(-(wave - self.center) ** 2 / (2. * sigma ** 2))

        # input line strength units in 10^-26 erg/cm^2/s (aka mJy)
        # because the flux is unitless and THIS calculation isn't unit-aware, we have to do some manual conversions
        int_flux = -integ.simps(flux * 1e-26, cs.c / (wave.to(u.m)))


        flux = flux * self.strength / int_flux

        return flux

    def optical_depth(self, wave):
        """
        Calculate absorption line optical depth over wavelength set, wave.

        Arguments
        ---------
        wave: np.ndarray
            Wavelengths over which to calculate line optical depth

        Returns
        -------
        tau: np.ndarray
            Line optical depth at wave
        """
        sigma = self.wave_width / 2.3548
        tau = np.exp(-(wave - self.center) ** 2 / (2. * sigma ** 2))
        tau *= self.strength
        # tau is unitless, but the calculation will produce mJy because of self.strength
        return tau.value


class ConvolvedSceneCube(object):

    """
    The SceneCube contains the source flux distribution as seen through the optics
    of a telescope/instrument combination. It takes the sampled ModelSceneCube and
    convolves it with the appropriate PSF from PSFLibrary.

    This is a central class of the ETC. The cube has two spatial and one wavelength dimension.

    Parameters
    ----------
    Sources : list
        A list of Source instances containing the physical parameters
        of the sources within the scene.
    instrument : Instrument class
        An instance of the instrument class
    background : Background, optional
        A background spectrum.
    webapp: bool
        Determines whether extra API checks are done
    validate: bool
        Keyword controlling whether intermediate products are saved for validation.
    empty_scene : bool
        Set to True if the scene is empty, and several computationally intensive steps can be bypassed

    Attributes
    ----------
    PSFLibrary :
    instrument :
    Grid :
    flux_cube :
    flux_plus_bg :

    """

    def __init__(self, scene, instrument, background=None, webapp=False, validate=False, empty_scene=False):
        self.warnings = {}
        self.scene = scene
        if instrument.get_slit_pars()["slit_axis"] == "x":
            self.aper_width = instrument.get_slit_pars()['disp']
            self.aper_height = instrument.get_slit_pars()['xdisp']
        else:
            self.aper_height = instrument.get_slit_pars()['disp']
            self.aper_width = instrument.get_slit_pars()['xdisp']
        self.multishutter = instrument.get_slit_pars()['multishutter']
        nslice_str = instrument.get_slit_pars()['nslice']
        self.slit_shape = instrument.get_slit_pars()['slit_shape']

        if nslice_str is not None:
            self.nslice = int(nslice_str)
        else:
            self.nslice = 1

        self.empty_scene = empty_scene
        self.instrument = instrument
        if validate:
            self.psf_library = instrument.psf_library
            self.background = background

        self.fov_size = self.get_fov_size()

        # Figure out what the relevant wavelength range is, given the instrument mode
        wrange = self.instrument.get_wave_range()

        self.source_spectra = []
        spectra = []

        # run through the sources and check their wavelength extents. warn if they fall short of the
        # current instrument configuration's range.
        mins = []
        maxes = []
        key = None
        for i, src in enumerate(scene.sources):
            spectrum = AstroSpectrum(src, webapp=webapp)
            self.warnings.update(spectrum.warnings)
            smin = spectrum.wave.min()
            smax = spectrum.wave.max()
            if smin > wrange['wmin']:
                if smin > wrange['wmax']:
                    key = "spectrum_missing_red"
                    msg = warning_messages[key] % (smin, smax, wrange['wmax'])
                    self.warnings["%s_%s" % (key, i)] = msg
                else:
                    key = "wavelength_truncated_blue"
                    msg = warning_messages[key] % (smin, wrange['wmin'])
                    self.warnings["%s_%s" % (key, i)] = msg
            if smax < wrange['wmax']:
                if smax < wrange['wmin']:
                    key = "spectrum_missing_blue"
                    msg = warning_messages[key] % (smin, smax, wrange['wmin'])
                    self.warnings["%s_%s" % (key, i)] = msg
                else:
                    key = "wavelength_truncated_red"
                    msg = warning_messages[key] % (smax, wrange['wmax'])
                    self.warnings["%s_%s" % (key, i)] = msg

            spectra.append(spectrum)

            mins.append(smin)
            maxes.append(smax)

        wmin = max([np.array(mins).min(), wrange['wmin']])
        wmax = min([np.array(maxes).max(), wrange['wmax']])

        # make sure we have something within range and error out otherwise
        if wmax < wrange['wmin'] or wmin > wrange['wmax']:
            msg = "No wavelength overlap between source_spectra [%.2f, %.2f] and instrument [%.2f, %.2f]." % (
                np.array(mins).min(),
                np.array(maxes).max(),
                wrange['wmin'],
                wrange['wmax']
            )
            raise RangeError(value=msg)

        # warn if partial overlap between combined wavelength range of all sources and the instrument's wrange
        if wmin != wrange['wmin'] or wmax != wrange['wmax']:
            key = "scene_range_truncated"
            self.warnings[key] = warning_messages[key] % (wmin, wmax, wrange['wmin'], wrange['wmax'])

        """
        Trim spectrum and do the spectral convolution here on a per-spectrum basis.  Most efficient to do it here
        before the wavelength sets are merged.  Also easier and much more efficient than convolving
        an axis of a 3D cube.
        Also compute the filter leak (only valid for image projections), which needs pre- and 
        post-trimming products.
        """
        for i, spectrum in enumerate(spectra):
            if self.instrument.projection_type in ["image", "image_scan"]:
                # get the pre-trimming filter leak components
                full_filtered = self.focal_plane_rate(spectrum.flux, spectrum.wave)
                full = integ.simps(full_filtered, spectrum.wave)

            # we trim here as an optimization so that we only convolve the section we need of a possibly very large spectrum
            spectrum.trim(wrange['wmin'], wrange['wmax'])

            if self.instrument.projection_type in ["image", "image_scan"]:
                # get the post-trimming filter leak components
                trim_filtered = self.focal_plane_rate(spectrum.flux, spectrum.wave)
                trim = integ.simps(trim_filtered, spectrum.wave)

                # compute filter leak and set warning if applicable
                filter_leak = (full-trim)/full
                if filter_leak > self.instrument.max_filter_leak:
                    key = "filter_leak"
                    msg = warning_messages[key].format(filter_leak*100)
                    self.warnings["%s_%s" % (key, i)] = msg

            spectrum = instrument.spectrometer_convolve(spectrum)
            self.source_spectra.append(spectrum)
        # include the background in the wavelength array calculation, so it's eventually properly sampled
        background.trim(wrange['wmin'], wrange['wmax'])

        """
        different spectra will have different sets of wavelengths. the obvious future
        case will be user-supplied spectra, but this is also true for analytic spectra
        that have different emission/absorption lines. go through each of the spectra,
        merge all of the wavelengths sets into one, and then resample each spectrum
        onto the combined wavelength set.
        """
        self.wave = self.source_spectra[0].wave
        for s in self.source_spectra:
            # Specifying the threshold fixes bug JETC-1758
            self.wave = syn.utils.merge_wavelengths(self.wave, s.wave, threshold=PANDEIA_ATOL)
        # Also merge in the background wavelengths
        self.wave = syn.utils.merge_wavelengths(self.wave, background.wave, threshold=PANDEIA_ATOL)

        projection_type = instrument.projection_type

        """
        For the spectral projections, we could use the pixel sampling. However, this
        may oversample the cube for input spectra with no narrow features. So we first check
        whether the pixel sampling will give us a speed advantage. Otherwise, do not resample to
        an unnecessarily fine wavelength grid.
        """
        if projection_type in ('spec', 'slitless', 'slitless_scan', 'multiorder'):
            wave_pix = instrument.get_wave_pix()
            wave_pix_trim = wave_pix[np.where(np.logical_and(wave_pix >= wrange['wmin'],
                                                             wave_pix <= wrange['wmax']))]
            if wave_pix_trim.size < self.wave.size:
                self.wave = wave_pix_trim

        """
        There is no inherently optimal sampling for imaging modes, but we resample here to
        a reasonable number of wavelength bins if necessary. This helps keep the cube rendering reasonable
        for large input spectra. Note that the spectrum resampling now uses synphot's flux conserving method
        """
        if projection_type in ('image', 'image_scan'):
            if self.wave.size > SPECTRAL_MAX_SAMPLES:
                self.wave = np.linspace(wrange['wmin'], wrange['wmax'], SPECTRAL_MAX_SAMPLES)

        self.nw = self.wave.size
        self.total_flux = np.zeros(self.nw)

        for spectrum in self.source_spectra:
            nyflux = np.nan_to_num(spectrum_resample(spectrum.flux, spectrum.wave, self.wave))
            spectrum.wave = self.wave
            spectrum.flux = nyflux
            spectrum.nw = self.nw
            self.total_flux += spectrum.flux

        # also need to resample the background spectrum
        if background is not None:
            background.resample(self.wave)

        self.grid, self.aperture_list, self.flux_cube_list, self.flux_plus_bg_list = \
            self.create_flux_cube(background=background)

        if validate:
            self.dist = self.grid.dist()

    def brightest_pixel_rate(self, spectrum, central_profile_fraction, unique_offset):
        """
        Calculate the HST Brightest Pixel Rate in a pixel-less way.
        
        This only works for imaging projection types. It is assumed the calling method 
        knows that information.

        1. determine the fraction of total flux that lands on the brightest pixel (for 
           point sources).
        2. determine the fraction of that flux that lands on the detector (scale for the
           fact that the PSFs don't sum to 1)
        3. determine the fraction of flux that lands in the central pixel of the 
           (possibly extended) profile
        4. determine the fcal plane rate of the spectrum and integrate to a scalar value

        Multiplying the results of 1-4 together yields the brightest pixel rate (without dark current and postflash)

        5. add in dark current, postflash, and background. This happens in 
           DetectorSignal.calculate_brightest_pixel()

        Parameters:
        -----------
        spectrum: AstroSpectrum
            The spectrum of a single source
        central_profile_fraction: float
            The fraction of flux that ends up in the center pixel of the profile, 
            computed by the Profile class
        
        """
        # 1. Due to optical abberations, not all PSFs are brightest at the center.
        #    Correction factors have been derived that will compute the flux at
        #    various offsets relative to the total flux on the detector, as a 
        #    function of wavelength. Loop through them all to find the one that,
        #    when paired with this spectrum, produces the brightest pixel.
        fluxfrac = True
        try:
            ee_filename = default_refdata_directory + "/" + self.instrument.telescope.tel_name + "/" + self.instrument.inst_name + f"/optical/{self.instrument.inst_name}_ee_curves.fits"

            # The file contains a whole bunch of centering offsets that have
            # effectively variable brightness. Iterate through all of them to
            # find the one that would result in the greatest multiplier.
            with fits.open(ee_filename) as hdu:
                table_name = self.instrument.mode.upper()
                t = Table(hdu[table_name].data)
                fluxfrac = hdu[table_name].header["FLUXFRAC"]

                corr_wave = t['wavelength']/10000 # The file's units are Angstroms

                # Remove the column containing only wavelengths to just leave the aperture sizes in the table.
                t.remove_column('wavelength')

            corr_bright = []
            corrected_rates = []
            for offset in t.columns:
                # As of 2023-07-12, this portion of the code is only trivially used to check single
                # curves. If this functionality is expanded this code needs to be tested more thoroughly
                # tested as the rest of this module.
                if len(t.columns) > 1:
                    self.warnings["untested_code"] = "The FLUXFRAC path in the brightest_pixel_rate code has not been well tested."
                corr_interp = interp.make_interp_spline(corr_wave, t[offset], k=1)
                corr_offset = corr_interp(self.wave)

                corr_rate = spectrum.flux * corr_offset

                corr_bright.append(integ.trapezoid(corr_rate, self.wave))
                corrected_rates.append(corr_rate)
            max_val = np.argmax(corr_bright)
            corr_pix_rate = corrected_rates[max_val]
        except FileNotFoundError as exp:
            corr_pix_rate = spectrum.flux


        # 2. determine the fraction of that flux that lands on the detector, from the
        #    PSF. This is pre-tabulated in the PSFs and exposed via self.psf_library
        if fluxfrac:
            # As of 2023-07-12, none of our data files use this data path; it is not as well
            # tested as the rest of this module.
            self.warnings["untested_code"] = "The FLUXFRAC path in the brightest_pixel_rate code has not been well tested."
            centered = [psf for psf in self.instrument.psf_library._psfs if psf["source_offset"] == unique_offset]
            psf_wave, psf_sum = np.asarray([[x["wave"] for x in centered],[x["psfsum"] for x in centered]])

            # interpolate onto this observation's wavelength scale
            psf_interp = interp.make_interp_spline(psf_wave, psf_sum, k=1)
            psf_sum = psf_interp(self.wave)
        else:
            psf_sum = 1

        corr_psf_pix_rate = psf_sum * corr_pix_rate


        # 3. Determine the fraction of the flux that lands in the central pixel of
        #    the profile. This has been done for us by ModelSceneCube.
        #    It may be higher than 1 if the extended source was normalized to a point
        #    somewhere far from the center.
        corr_psf_profile_pix_rate = corr_psf_pix_rate * central_profile_fraction


        # 4. determine the focal plane rate of the spectrum and integrate to a scalar
        #    value 
        # Get the quantum yield
        corr_psf_profile_fp_pix_rate = self.focal_plane_rate(self.ote_rate(corr_psf_profile_pix_rate), spectrum.wave)
        q_yield, fano_factor = self.instrument.get_quantum_yield(self.wave)
        # convert the photon rate to electron rate by multiplying by the quantum
        # yield which is a function of wavelength
        brightest_pixel_rate = integ.trapezoid(corr_psf_profile_fp_pix_rate * q_yield, self.wave)
        # brightest_pixel_rate is now the rate (in e-/s or counts/s) of the source, in its
        # brightest pixel.

        return brightest_pixel_rate


    def get_fov_size(self, pixbuffer=SCENE_FOV_PIXBUFFER):
        # The scene size is the minimum size containing all sources, but at least as large as the PSF.


        instrument_name = self.instrument.get_name()
        aperture_name = self.instrument.get_aperture()
        readout_pattern = self.instrument.detector.get("readout_pattern", "default")
        psf_shape = self.instrument.psf_library.get_shape(instrument_name, aperture_name)
        psf_pix_scl_x, psf_pix_scl_y = self.instrument.psf_library.get_pix_scale(instrument_name, aperture_name)
        psf_size = np.max([psf_shape[1] * psf_pix_scl_x, psf_shape[0] * psf_pix_scl_y])

        if self.instrument.projection_type == 'multiorder':
            # for the multiorder case (i.e. SOSS), we need to set the FOV size based on the subarray used
            subarray = self.instrument.detector['subarray']
            aperture = self.instrument.instrument['aperture']
            subarray_config = self.instrument.subarray_config['default'][subarray]
            if readout_pattern in self.instrument.subarray_config:
                subarray_config = self.instrument.subarray_config[readout_pattern][subarray]
            nx = subarray_config['nx']
            ny = subarray_config['ny']
            pix_scale_x, pix_scale_y = self.instrument.psf_library.get_det_scale(instrument_name, aperture_name)
            disp_axis = self.instrument.dispersion_axis()
            if disp_axis == "x":
                fov_pix = ny
                fov_size = fov_pix * pix_scale_x
            else:
                fov_pix = nx
                fov_size = fov_pix * pix_scale_y

        elif self.instrument.dynamic_scene:
            scene_size = self.scene.get_size()
            # a scene size and maximum scene size must be defined for each instrument/mode.
            # if they're not defined at all, it's a data problem.
            try:
                inst_fov = self.instrument.scene_size
                max_fov = self.instrument.max_scene_size
            except AttributeError as e:
                message = "Instrument configuration must specify default and maximum scene sizes. (%s)" % e
                raise DataError(value=message)
            # the configured scene size must be smaller than the maximum. since they can come from
            # either the data files or input configuration, this is an input error.
            if inst_fov > max_fov:
                message = "Specified scene size, %f arcsec, larger than maximum allowed size of %f arcsec" % (inst_fov, max_fov)
                raise EngineInputError(value=message)

            # find the largest of the size of the defined scene, the configured instrument FOV, and the PSF image size.
            # if this is larger than the configured maximum FOV size, set it to the maximum value.
            fov_size = np.max([scene_size + pixbuffer * np.max([psf_pix_scl_x, psf_pix_scl_y]), inst_fov, psf_size])
            # warn if the fov size is larger than the instrument's maximum fov size.
            if fov_size > max_fov:
                key = "max_scene_size_reached"
                self.warnings[key] = warning_messages[key] % (fov_size, max_fov)
                fov_size = max_fov
        else:
            # If the scene is not dynamic and not multiorder, there are still cases where we want
            # to customize the scene size, like NIRSpec IFU (JETC-1704) and STIS CCD ImagingAcq 
            # (JETC-1584)
            try:
                fov_size = self.instrument.custom_scene_size
            except AttributeError as e:
                # otherwise, the usual behavior: Scene size is PSF size
                fov_size = psf_size

        # check to make sure at least one source is within the field of view. warn otherwise...
        if fov_size < self.scene.get_min_size():
            key = "scene_fov_too_small"
            self.warnings[key] = warning_messages[key] % fov_size

        return fov_size

    def focal_plane_rate(self, rate, wave):
        """
        Takes the output from self.ote_rate() and multiplies it by the components of efficiency within the
        system and returns the source rate at the focal plane in e-/s/pixel/micron.
        """
        filter_eff = self.instrument.get_filter_eff(wave)
        disperser_eff = self.instrument.get_disperser_eff(wave)
        internal_eff = self.instrument.get_internal_eff(wave)
        qe = self.instrument.get_detector_qe(wave)

        fp_rate = rate * filter_eff * disperser_eff * internal_eff * qe

        return fp_rate
    
    def ote_rate(self, rate):
        """
        Calculate source rate in e-/s/pixel/micron at the telescope entrance aperture given a flux cube in
        mJy/pixel.
        """
        # spectrum in mJy/pixel, wave in micron, f_lambda in photons/cm^2/s/micron
        f_lambda = CONVERSION_CONSTANT * (rate / np.array(self.wave, dtype=np.float32))
        ote_int = np.single(self.instrument.telescope.get_ote_eff(self.wave))
        coll_area = np.single(self.instrument.telescope.coll_area)
        a_lambda = coll_area * ote_int
        # e-/s/pixel/micron
        ote_rate = np.array(f_lambda * a_lambda, dtype=np.float32)

        return ote_rate

    def create_flux_cube(self, background=None):
        """
        Generate the list of convolved flux cubes that will go into the ETC calculation.
        The spectral convolution is already done and the spatial convolution is performed here
        using self.PSFLibrary (nominally as generated by webbPSF). The flux cube handles position-dependent
        PSFs by assigning PSF profiles to individual sources, convolving intermediate cubes of sources with
        common PSFs and finally co-adding cubes to create a final master flux cube. Typical ETC calculations, which do
        not have position-dependent PSFs, are not affected by this functionality.

        Parameters
        ----------
        background: background.Background instance

        Returns
        -------
        <tuple>:
            spatial grid used to create cube(s) (coords.Grid instance)
            list of apertures (list)
            list of flux cubes (list; one per aperture)
            list of flux cubes including background (list; one per aperture)

        """
        if self.instrument.psf_library is None:
            raise NotImplementedError('The use of a simple PSF in isolation is deprecated. Please provide PSFLibrary.')

        ### 1. Load information about PSFs and scene
        instrument_name = self.instrument.get_name()
        aperture_name = self.instrument.get_aperture()
        psf_pixsize_x, psf_pixsize_y = self.instrument.psf_library.get_pix_scale(instrument_name, aperture_name)
        psf_upsamp = self.instrument.psf_library.get_upsamp(instrument_name, aperture_name)
        kernel_npix_x, kernel_npix_y = self.instrument.psf_library.get_shape(instrument_name, aperture_name)

        detector_npix_x = int(np.round(self.fov_size / psf_pixsize_x / psf_upsamp))
        detector_npix_y = int(np.round(self.fov_size / psf_pixsize_y / psf_upsamp))

        if detector_npix_x % 2 == 0:
            detector_npix_x += 1
        if detector_npix_y % 2 == 0:
            detector_npix_y += 1
        detector_shape = (detector_npix_y, detector_npix_x)

        ### 2. Create the cubes to fill
        # The cube has to be in a y,x orientation
        # see coord.py Grid definition for more explanation
        flux_cube_list = [
            np.zeros(
                (int(detector_shape[0]),
                 int(detector_shape[1]),
                 self.nw), dtype=np.float32) for ir in range(self.nslice)]

        flux_plus_bg_list = [
            np.zeros(
                (int(detector_shape[0]),
                 int(detector_shape[1]),
                 self.nw), dtype=np.float32) for ir in range(self.nslice)]

        ### 3. Get background
        if background is not None:
            pupil_thru = self.instrument.psf_library.get_pupil_throughput(self.wave[0], instrument_name, aperture_name)
            self.bg = background.mjy_pix * pupil_thru
        else:
            self.bg = self.wave * 0.0

        ### 4. Create Model Scene Cube(s) from scene and source information
        self.grid = Grid(psf_pixsize_x * psf_upsamp, psf_pixsize_y * psf_upsamp, detector_shape[1], detector_shape[0])
        fine_grid = Grid(psf_pixsize_x, psf_pixsize_y, detector_shape[1] * psf_upsamp, detector_shape[0] * psf_upsamp)
        psf_associations = self.instrument.psf_library.associate_offset_to_source(self.scene.sources, instrument_name, aperture_name)
        unique_offsets = list(set(psf_associations))

        current_scenes = []
        for unique_offset in unique_offsets:
            offset_indices = [i for (i, v) in enumerate(psf_associations) if v == unique_offset]
            current_scene = ModelSceneCube([self.source_spectra[i] for i in offset_indices], fine_grid, psf_upsamp)
            # Catch calculations that are too bright to be computed, and have them abort immediately.
            if not np.all(np.isfinite(current_scene.int)):
                # warn the user about what they've done
                raise EngineInputError(value="Calculation error: at least one source is too bright to be computed "
                                             "correctly. Check your sources and source normalizations.")
            current_scenes.append(current_scene)

            # Do this part within the unique_offset loop so we match the correct profiles to the correct spectra
            if "imag" in self.instrument.projection_type and self.instrument.telescope.tel_name == "hst":
                brightest_pixel_list = []
                for idx, spectrum in enumerate([self.source_spectra[i] for i in offset_indices]):
                    brightest_pixel_list.append(self.brightest_pixel_rate(spectrum, current_scene.central_profiles_list[idx], unique_offset))
                self.brightest_pixel_rate_imaging = np.max(brightest_pixel_list)

        ### 5. Speedup check: Is this a single centered point source?
        # Check whether we have only a single point source near the center
        # (if we do, the convolution can be faster because we don't have to convolve a field larger
        # than the PSF kernel size, even if the scene is formally larger)
        self.single_point_source = True
        self.single_centered_source = True if len(self.scene.sources) == 1 else False # useful elsewhere.
        for src in self.scene.sources:
            if np.abs(src.position['x_offset']) > psf_pixsize_x*10:
                self.single_point_source = False
                self.single_centered_source = False
            elif np.abs(src.position['y_offset']) > psf_pixsize_y*10:
                self.single_point_source = False
                self.single_centered_source = False
            elif src.shape['geometry'] != 'point':
                self.single_point_source = False

        ### 6. Check and prepare scene parameters
        scene_npix_x = detector_shape[1] * psf_upsamp
        scene_npix_y = detector_shape[0] * psf_upsamp
        npix_x = int(detector_shape[1])
        npix_y = int(detector_shape[0])
        # Check to make sure npix is an integer
        if not (scene_npix_x % psf_upsamp) == 0:
            # This should never happen
            message = "Number of scene pixels, {}, is not divisible by the PSF upsampling factor, {}".format(
                scene_npix_x, psf_upsamp)
            raise DataConfigurationError(value=message)
        # Check to make sure npix is an integer
        if not (scene_npix_y % psf_upsamp) == 0:
            # This should never happen
            message = "Number of scene pixels, {}, is not divisible by the PSF upsampling factor, {}".format(
                scene_npix_y, psf_upsamp)
            raise DataConfigurationError(value=message)
        # close enough?
        # TODO: Use numpy.isclose with our usual parameters?
        if (np.abs(current_scene.xsamp / psf_pixsize_x - 1) > 1e-10):
            raise ValueError("scene sampling must be the same as PSF sampling")
        if (np.abs(current_scene.ysamp / psf_pixsize_y - 1) > 1e-10):
            raise ValueError("scene sampling must be the same as PSF sampling")
        # We have to make sure npix is an integer or it can't be used as an index
        new_shape = (npix_y, npix_x)

        ### 7. Make PSF convolution window parameters
        mini_x = int((scene_npix_x - kernel_npix_x)/2) # minimum index of the kernel size within the FOV
        maxi_x = int((scene_npix_x + kernel_npix_x)/2) # maximum index of the kernel size within the FOV
        mini_y = int((scene_npix_y - kernel_npix_y)/2) # minimum index of the kernel size within the FOV
        maxi_y = int((scene_npix_y + kernel_npix_y)/2) # maximum index of the kernel size within the FOV

        # if we have a rare instance where the PSF is larger than the scene, just effectively convolve it normally
        # (without this, the tiny window would actually have no pixels in it, producing a zero cube)
        if kernel_npix_x > scene_npix_x:
            mini_x = 0
            maxi_x = scene_npix_x
        if kernel_npix_y > scene_npix_y:
            mini_y = 0
            maxi_y = scene_npix_y

        ### 8. Make the masks
        fine_masks = []
        masks = []
        if self.aper_width is not None and self.aper_height is not None and self.slit_shape=="rectangle":
            offsets = [(i - (self.nslice - 1) / 2.) * self.aper_width for i in np.arange(self.nslice)]
            aperture_list = []
            for offset in offsets:
                slice_mask_fine = np.zeros((scene_npix_y,scene_npix_x))
                # Is this a multishutter instrument?
                if self.multishutter:
                    for shutter in self.multishutter:
                        new_mask = fine_grid.rectangular_mask(
                            width=self.aper_width,
                            height=self.aper_height,
                            xoff=offset + shutter[0],
                            yoff=shutter[1]
                        )
                        slice_mask_fine = np.maximum(slice_mask_fine, new_mask)
                else:
                    slice_mask_fine = fine_grid.rectangular_mask(
                        width=self.aper_width,
                        height=self.aper_height,
                        xoff=offset,
                        yoff=0.0
                    )
                fine_masks.append(slice_mask_fine)
                masks.append(AdvancedPSF._rebin(self,slice_mask_fine, new_shape))
                aperture_list.append({'width': self.aper_width, 'height': self.aper_height, 'offset': (0., offset)})
        elif self.aper_width is not None and self.slit_shape=="circle":
            # we do not have a use case for multiple or non-centered circular slits
            slice_mask_fine = fine_grid.circular_mask(
                self.aper_width/2, 
                xoff=0.0, 
                yoff=0.0
            )
            fine_masks.append(slice_mask_fine)
            masks.append(AdvancedPSF._rebin(self,slice_mask_fine, new_shape))
            aperture_list = [{'radius': self.aper_width, 'offset': (0., 0)}]
        else:
            # The cube indexing order is y,x
            slice_mask_fine = np.ones((scene_npix_y,scene_npix_x))
            fine_masks.append(slice_mask_fine)
            masks.append(AdvancedPSF._rebin(self,slice_mask_fine, new_shape))
            aperture_list = [self.grid.get_aperture()]


        ### 9. Run the convolution for each wavelength, for each unique PSF offset
        for iw in np.arange(self.nw):
            for current_scene, unique_offset, i in zip(current_scenes, unique_offsets, list(range(len(unique_offsets)))):
                profile = self.instrument.psf_library.get_psf(self.wave[iw], instrument_name, aperture_name, source_offset=unique_offset)
                psf_upsamp = profile['upsamp']
                kernel = np.asarray(profile['int'], dtype=np.float32)

                if i == 0:
                    psf = AdvancedPSF(
                        self.wave[iw],
                        kernel,
                        psf_upsamp=psf_upsamp,
                        current_scene=current_scene,
                        fine_masks=fine_masks,
                        masks=masks,
                        fine_shape=(scene_npix_y, scene_npix_x),
                        new_shape=new_shape,
                        grid=self.grid,
                        aperture_list=aperture_list,
                        mini_x=mini_x,
                        maxi_x=maxi_x,
                        mini_y=mini_y,
                        maxi_y=maxi_y,
                        bg_w=self.bg[iw],
                        single=self.single_point_source,
                        empty_scene = self.empty_scene
                    )
                    self.warnings.update(psf.warnings)
                else:
                    # if there are sources with different PSFs, calculate their intensities and add them,
                    # but do not add more background.
                    psf.add_intensity(
                        AdvancedPSF(
                            self.wave[iw],
                            kernel,
                            psf_upsamp=psf_upsamp,
                            current_scene=current_scene,
                            fine_masks=fine_masks,
                            masks=masks,
                            fine_shape=(scene_npix_x, scene_npix_y),
                            new_shape=new_shape,
                            grid=self.grid,
                            aperture_list=aperture_list,
                            mini_x=mini_x,
                            maxi_x=maxi_x,
                            mini_y=mini_y,
                            maxi_y=maxi_y,
                            bg_w=0.
                        )
                    )
                    self.warnings.update(psf.warnings)

            for islice in range(self.nslice):
                flux_cube_list[islice][:, :, iw] = psf.slice_int_list[islice]
                flux_plus_bg_list[islice][:, :, iw] = psf.slice_int_plus_bg_list[islice]

        return self.grid, aperture_list, flux_cube_list, flux_plus_bg_list

    def spectral_model_transform(self):
        """
        Create engine API format dict section containing properties of the wavelength coordinates
        used in the construction of a ConvolvedSceneCube.

        Returns
        -------
        t: dict (engine API compliant keys)
        """
        t = {}
        t['wave_refpix'] = 0
        t['wave_refval'] = self.wave[0]
        t['wave_max'] = self.wave.max()
        t['wave_min'] = self.wave.min()
        t['wave_size'] = self.wave.size
        # this is a bit of a hack since the wavelength sampling is NOT constant at this stage.
        # this is simply the mean step and should probably be removed altogether. if one wishes to
        # plot anything from this stage, they should instead use self.wave which provides the true
        # mapping of index -> wavelength
        if len(self.wave > 1):
            t['wave_step'] = (self.wave[-1] - self.wave[0]) / self.wave.size
        else:
            t['wave_step'] = 0.0
        return t

    def cube_wcs_info(self):
        """
        Create WCS headers and FITS binary table that describe the cube's coordinate system.

        Returns
        -------
        tbhdu: astropy.io.fits.BinTableHDU instance
            The wavelength sampling is irregularly spaced so we define a binary FITS
            table that contains the array of wavelengths, self.wave.
        header: dict
            Contains the WCS keys that define the coordinate transformation for all axes
        """
        header = self.grid.wcs_info()
        header['ctype3'] = 'WAVE-TAB'
        header['cname3'] = 'Wavelength'
        header['cunit3'] = 'um'
        header['PS3_0'] = 'WCS-TAB'
        header['PS3_1'] = 'WAVELENGTH'
        header['PS3_2'] = 'WAVE-INDEX'
        tbhdu = fits.BinTableHDU.from_columns([
            fits.Column(name='WAVELENGTH',
                        unit='um',
                        format="1D",
                        array=self.wave),
            fits.Column(name='WAVE-INDEX',
                        format="1J",
                        array=np.arange(self.wave.size))
        ])
        tbhdu.name = 'WCS-TAB'
        return tbhdu, header

    def export_to_fits(self, fitsfile='ModelDetectorCube'):
        """
        Write convolved scene cube to a FITS file
        """
        header = self.grid.wcs_info()
        for slice_index,flux_plus_bg in enumerate(self.flux_plus_bg_list):
            fitsfile_slice = '{}{}.fits'.format(fitsfile, str(slice_index).strip())
            t1 = fits.PrimaryHDU(np.moveaxis(flux_plus_bg, [0, 1, 2], [1, 2, 0]))
            c1 = fits.Column(name='wavelength', array=self.wave, format='D')
            t2 = fits.BinTableHDU.from_columns([c1])
            t1.header.update(header)
            tbhdu = fits.HDUList([t1,t2])
            tbhdu.writeto(fitsfile_slice, overwrite=True)
            tbhdu.close()


class AdvancedPSF(object):

    """
    Convolve scene with PSF from a PSF library (e.g., one calculated using WebbPSF).

    Parameters
    ----------
    wave: float
        Wavelength (in microns) at which the scene is to be convolved.
    kernel: np.ndarray
        An interpolated PSF at the wavelength specified by wave.
    psf_upsamp: float
        The amount by which the PSF is oversampled
    current_scene: np.ndarray
        The plane of the scene we're operating on.
    fine_masks: np.ndarray
        A mask array showing geometric slit sizes (oversampled)
    masks: np.ndarray
        A mask array showing geometric slit sizes (non-oversampled)
    grid: Grid or IrregularGrid
        The grid describing the FOV, providing a mapping of pixel to position in arcseconds
    aperture_list: list
        A list of all the aperture masks (slits) that went into making the mask and fine_mask
    mini_x:
    maxi_x:
    mini_y:
    maxi_y:
        Precomputed parameters by which the scene might need to be sliced, for single-star mode.
    fine_shape: tuple
        The dimensions of the upsampled (fine) mask
    new_shape: tuple
        The dimensions of the non-upsampled mask
    bg_w: float
        Sky background rate at wavelength.
    single: Bool
        If True, assume that we only have a single point source near or at the center, and then
        only convolve a field the size of the PSF kernel (leaving the rest of the FOV at zero source flux).
    empty_scene: Bool
        If True, it's an empty scene and we can just return the empty array.
    """
    def __init__(self, wave, kernel, psf_upsamp=1, current_scene=None, fine_masks=None, masks=None, grid=None,
                 aperture_list=[], mini_x=0, maxi_x=-1, mini_y=0, maxi_y=-1, fine_shape=(0, 0), new_shape=(0, 0),
                 bg_w=0, single=False, empty_scene=False):

        # For API compatibility
        self.aperture_list = aperture_list
        self.grid = grid
        self.warnings = {}

        windex = np.abs(current_scene.wave - wave).argmin() # find nearest index

        self.slice_int_list = []
        self.slice_int_plus_bg_list = []
        self.slice_mask_list = []

        if empty_scene:
            self.intensity = np.zeros(fine_shape)
            # See Issue #71 for discussion and benchmarks about
            # these different FFT methods. astropy.convolution plus pyFFTW yields a significant
            # speed-up, of order 30% or more.
            #
            #The original method
            # import stsci.convolve as stsci
            # self.intensity = stsci.convolve2d(model_scene.int[:, :, windex], kernel, fft=1, mode='wrap')
            #
            # Basic astropy.convolution with FFTs with parameters set to produce numbers that
            # match the original method
            # self.intensity = convolve_fft(model_scene.int[:, :, windex], kernel[:-1, :-1], normalize_kernel=False)
        elif single:
            # This does not use the grid class, so indexing must be y,x (see the Grid class for more)
            self.intensity = np.zeros(fine_shape)
            # mode='same' forces the FFT output to be the same shape as the input scene to be convolved with
            # (first argument to the function)
            self.intensity[mini_y:maxi_y, mini_x:maxi_x] = scipy.signal.fftconvolve(current_scene.int[mini_y:maxi_y, mini_x:maxi_x, windex],kernel,mode='same')
        else:
            self.intensity = scipy.signal.fftconvolve(current_scene.int[:, :, windex], kernel, mode='same')

        # Add the background
        self.intensity_plus_bg = self.intensity + bg_w / psf_upsamp ** 2

        """
        We can now operate with any number of physical spectral apertures (slices) of the FOV. A single slit
        mode simply has nslice=1. An imaging mode is also a slice, but with infinite aperture.
        A multishutter instrument can create a slice aperture mask consisting of a discrete number of mutually
        offset rectangles. In principle, one could create an IFU with each slice consisting of discrete shutters.
        This could be used to simulate different IFU designs, such as lenslet or micro-mirror arrays.
        """
        for idx, fine_mask in enumerate(fine_masks):
            slice_int, slice_int_plus_bg = self._apply_slit_mask(fine_mask, new_shape)
            # for this purpose a set of multiple shutters is treated as a single aperture and uses
            # the properties of the central shutter.
            
            self.slice_int_list.append(slice_int)
            self.slice_int_plus_bg_list.append(slice_int_plus_bg)
            self.slice_mask_list.append(masks[idx])


    def add_intensity(self, psf):
        """
        Add another compatible PSF intensity.

        Parameters
        ----------
        psf : AdvancedPSF instance
            A previously computed AdvancedPSF to add.

        """

        self.slice_int_list = [slice_int + add_slice_int for slice_int, add_slice_int in
                               zip(self.slice_int_list, psf.slice_int_list)]
        self.slice_int_plus_bg_list = [slice_int + add_slice_int for slice_int, add_slice_int in
                                       zip(self.slice_int_plus_bg_list, psf.slice_int_plus_bg_list)]

        self.warnings.update(psf.warnings)

    def _apply_slit_mask(self, slit_mask, new_shape):
        """
        Applies a slit mask, and also rebins to a new shape.

        Parameters
        ----------
        slit_mask : 2D ndarray
            Slit mask as generated by self._create_slit_mask()
        new_shape : list-like
            New shape for array after binning

        Returns: list-like of 2D ndarrays
            Intensity image, intensity plus background image, aperture mask
        """
        intensity = self._rebin(self.intensity * slit_mask, new_shape)
        intensity_plus_bg = self._rebin(self.intensity_plus_bg * slit_mask, new_shape)
        return intensity, intensity_plus_bg

    def _rebin(self, a, shape):
        """
        Re-bin a 2D array.

        Parameters
        ----------
        a : ndarray
            Array to be re-binned
        shape : list-like
            New shape after re-binning

        Returns
        -------
        ndarray
        """
        # While the array shape is defined as x,y, it's actually implemented as y,x; hence the subscripts
        # see coord.py Grid definition for more explanation
        sh = int(shape[0]), int(a.shape[0] // shape[0]), int(shape[1]), int(a.shape[1] // shape[1])
        new = a.reshape(sh).sum(-1).sum(1)
        return new

    def _rebin_1d_mean(self, a, shape):
        """
        Re-bin a 1D array using averaging.

        Parameters
        ----------
        a : 1D ndarray
            Array to be re-binned
        shape : int
            New length after re-binning

        Returns
        -------
        ndarray
        """
        sh = shape, a.shape[1] // shape
        new = a.reshape(sh).mean(-1)
        return new



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .custom_exceptions import EngineInputError
from .utils import recursive_subclasses, merge_data
from .instrument import Instrument

# need to import the Instrument subclasses from whereever they're defined
from .jwst import NIRSpec, NIRCam, NIRISS, MIRI
from .roman import WFI
from .hst import WFC3, ACS, STIS, COS

def InstrumentFactory(config={}, webapp=False, **kwargs):
    """
    Function to take configuration data and build/return an appropriately configured
    instance of the desired Instrument subclass.

    Parameters
    ----------
    config: dict
        Configuration data in engine API format.  Used to suss out which instrument to configure
        and then passed along.
    **kwargs: keyword/value pairs
        Additional configuration data
    """
    all_config = merge_data(config, dict(**kwargs))
    types = recursive_subclasses(Instrument)
    instruments = [t.__name__.lower() for t in types]
    inst_map = dict(list(zip(instruments, types)))

    # get the instrument name out of the input configuration
    try:
        instrument = all_config['instrument']['instrument']
    except KeyError as e:
        msg = "Must provide Instrument name via engine API compatible dict or instrument=<name> keyword/value pair."
        raise EngineInputError(value=msg)

    # get the instrument mode out of the input configuration
    try:
        mode = all_config['instrument']['mode']
    except KeyError as e:
        msg = "Must provide Instrument mode via engine API compatible dict or mode=<modename> keyword/value pair. (%s)" % e
        raise EngineInputError(value=msg)

    # make sure requested instrument is one we actually support
    if instrument not in instruments:
        msg = "Instrument %s not supported/implemented." % instrument
        raise EngineInputError(value=msg)
    else:
        cls = inst_map[instrument](mode=mode, config=config, webapp=webapp, **kwargs)
        return cls



""" ACS-specific functions

"""



from ....engine import initialize


def get_configuration_data(ei):
    """ Load instrument and detector data from config files.

        This overrides the base function in order to handle
        ACS specifics.

    Parameters
    ----------
    ei: dict
      engine input

    Returns
    -------
    iconfig: dict
      dictionary of instrument data
    dconfig: dict
      dictionary of detector data
    custom_config: dict
      dictionary with custom data

    """
    # get baseline configuration data.
    iconfig, dconfig, custom_config = initialize.get_configuration_data(ei)

    # spectral resolution must be returned in the detector config dict.
    dconfig['spectral_resolution'] = iconfig[dconfig['detector_name']]['spectral_resolution']

    return iconfig, dconfig, custom_config




# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os

from . import io_utils as io
from . import config as cf
from .config import DefaultConfig
from .io_utils import ref_data_interp, ref_data_column
from .pandeia_warnings import telescope_warning_messages as warning_messages

default_refdata_directory = cf.default_refdata_directory


class TelescopeConfig(DefaultConfig):

    """
    This is a trivial top-level class for now. It overrides DefaultConfig._get_config() to get the configuration
    information from a different place, currently under $pandeia_refdata/<tel_name>/telescope.
    """

    def _get_config(self):
        """
        Read default configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # use this trick to key the configuration file name off the name of the instantiated subclass
        self.tel_name = self.__class__.__name__.lower()
        self.ref_dir = os.path.join(default_refdata_directory, self.tel_name, "telescope")
        config = io.read_json(os.path.join(self.ref_dir, "config.json"), raise_except=True)
        # add separate CR config, if it's there...
        cr_config = io.read_json(os.path.join(self.ref_dir, "cr_config.json"), raise_except=False)
        config.update(cr_config)
        return config


class Telescope(TelescopeConfig):

    """
    The telescope class contains information about the optical system and other effects
    that are common to all associated instruments.
    """

    def get_ote_eff(self, wave):
        """
        Get efficiency of the telescope optics, the OTE

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate OTE efficiency onto

        Returns
        -------
        ote: numpy.ndarray or float
            If file exists, return ote_efficiency(wave). Else return 1.0.
        """
        ote_file = os.path.join(self.ref_dir, self.paths['otefile'])
        if ote_file is not None:
            ote = ref_data_interp(ote_file, wave, colname='throughput')
        else:
            ote = 1.0
            key = "telescope_ote_efficiency_missing"
            self.warnings[key] = warning_messages[key]
        return ote

    def get_bg_tel(self):
        """
        Get the telescope-specific background

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate the telescope background onto

        Returns
        -------
        bg_tel: numpy.ndarray or float
            If file exists, return bg_tel(wave). Else return 0.0
        """
        bg_tel_file = os.path.join(self.ref_dir, self.paths['bg_tel'])
        if bg_tel_file is not None:
            bg_tel_wave = ref_data_column(bg_tel_file, colname='WAVELENGTH')
            bg_tel = ref_data_column(bg_tel_file, colname='SB')
        else:
            bg_tel_wave = [0.1, 30.0]
            bg_tel = [0.0, 0.0]
            key = "telescope_background_missing"
            self.warnings[key] = warning_messages[key]
        return bg_tel_wave, bg_tel
